"""IPSecure is an abstraction for handling a KNXnet/IP Secure layer."""

from __future__ import annotations

from abc import ABC, abstractmethod
import asyncio
from collections.abc import Callable
import logging
import random
from typing import Final

from cryptography.hazmat.primitives.asymmetric.x25519 import (
    X25519PrivateKey,
    X25519PublicKey,
)

from xknx.exceptions import (
    CommunicationError,
    CouldNotParseKNXIP,
    IPSecureError,
    KNXSecureValidationError,
)
from xknx.knxip import (
    HPAI,
    KNXIPFrame,
    KNXIPServiceType,
    RoutingIndication,
    SecureWrapper,
    SessionResponse,
    SessionStatus,
    TimerNotify,
)
from xknx.knxip.knxip_enum import SecureSessionStatusCode
from xknx.secure.security_primitives import (
    calculate_message_authentication_code_cbc,
    decrypt_ctr,
    derive_device_authentication_password,
    derive_user_password,
    encrypt_data_ctr,
    generate_ecdh_key_pair,
)
from xknx.secure.util import bytes_xor, sha256_hash
from xknx.util import asyncio_timeout

from .const import SESSION_KEEPALIVE_RATE, XKNX_SERIAL_NUMBER
from .request_response import Authenticate, Session
from .transport import KNXIPTransport, TCPTransport, UDPTransport

logger = logging.getLogger("xknx.log")
knx_logger = logging.getLogger("xknx.knx")
ip_secure_logger = logging.getLogger("xknx.ip_secure")


COUNTER_0_HANDSHAKE = (  # used in SessionResponse and SessionAuthenticate
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\x00"
)
MESSAGE_TAG_TUNNELLING = bytes.fromhex("00 00")  # use 0x00 0x00 for tunneling


class _IPSecureTransportLayer(ABC):
    """Abstract for Secure transport layer."""

    __slots__ = ()

    session_id: int
    _key: bytes

    @abstractmethod
    def get_sequence_information(self) -> bytes:
        """Return byte representation of sequence information."""

    @abstractmethod
    def get_message_tag(self) -> bytes:
        """Return byte representation of message tag."""

    def decrypt_frame(self, encrypted_frame: KNXIPFrame) -> KNXIPFrame:
        """Unwrap and verify KNX/IP frame from SecureWrapper."""
        # TODO: get raw data from KNXIPFrame class directly instead of recalculating it with to_knx()
        # TODO: refactor so assert isn't needed (maybe subclass SecureWrapper from KNXIPFrame instead of being an attribute)
        assert isinstance(encrypted_frame.body, SecureWrapper)
        if encrypted_frame.body.secure_session_id != self.session_id:
            raise KNXSecureValidationError("Invalid secure session id")

        session_id_bytes = encrypted_frame.body.secure_session_id.to_bytes(2, "big")
        wrapper_header = encrypted_frame.header.to_knx()

        dec_frame, mac_tr = decrypt_ctr(
            key=self._key,
            counter_0=(
                encrypted_frame.body.sequence_information
                + encrypted_frame.body.serial_number
                + encrypted_frame.body.message_tag
                + b"\xff\x00"
            ),
            mac=encrypted_frame.body.message_authentication_code,
            payload=encrypted_frame.body.encrypted_data,
        )
        mac_cbc = calculate_message_authentication_code_cbc(
            key=self._key,
            additional_data=wrapper_header + session_id_bytes,
            payload=dec_frame,
            block_0=(
                encrypted_frame.body.sequence_information
                + encrypted_frame.body.serial_number
                + encrypted_frame.body.message_tag
                + len(dec_frame).to_bytes(2, "big")
            ),
        )
        if mac_cbc != mac_tr:
            raise KNXSecureValidationError(
                "Verification of message authentication code failed"
            )

        knxipframe, _ = KNXIPFrame.from_knx(dec_frame)
        return knxipframe

    def encrypt_frame(self, plain_frame: KNXIPFrame) -> KNXIPFrame:
        """Wrap KNX/IP frame in SecureWrapper."""
        try:
            sequence_information = self.get_sequence_information()
        except OverflowError as err:
            raise IPSecureError(
                "KNX IP Secure sequence counter overflow."
                "\nCongratulations! You've managed to overflow a counter designed to last about 9000 years! "
                "Before celebrating, please check for any malfunctioning devices or suspicious activity "
                "in your installation. Once you've ensured everything's safe, reset the secure session "
                "to restore normal operation."
            ) from err
        message_tag = self.get_message_tag()
        plain_payload = plain_frame.to_knx()  # P
        payload_length = len(plain_payload)  # Q
        # 6 KNXnet/IP header, 2 session_id, 6 sequence_number, 6 serial_number, 2 message_tag, 16 MAC = 38
        total_length = 38 + payload_length
        # TODO: get header data and total_length from SecureWrapper class
        wrapper_header = bytes.fromhex("06 10 09 50") + total_length.to_bytes(2, "big")

        mac_cbc = calculate_message_authentication_code_cbc(
            key=self._key,
            additional_data=wrapper_header + self.session_id.to_bytes(2, "big"),
            payload=plain_payload,
            block_0=(
                sequence_information
                + XKNX_SERIAL_NUMBER
                + message_tag
                + payload_length.to_bytes(2, "big")
            ),
        )
        encrypted_data, mac = encrypt_data_ctr(
            key=self._key,
            counter_0=(
                sequence_information + XKNX_SERIAL_NUMBER + message_tag + b"\xff\x00"
            ),
            mac_cbc=mac_cbc,
            payload=plain_payload,
        )
        return KNXIPFrame.init_from_body(
            SecureWrapper(
                secure_session_id=self.session_id,
                sequence_information=sequence_information,
                serial_number=XKNX_SERIAL_NUMBER,
                message_tag=message_tag,
                encrypted_data=encrypted_data,
                message_authentication_code=mac,
            )
        )


class SecureSession(TCPTransport, _IPSecureTransportLayer):
    """Class for handling a KNXnet/IP Secure tunnelling session."""

    __slots__ = (
        "_device_authentication_code",
        "_keepalive_task",
        "_key",
        "_peer_public_key",
        "_private_key",
        "_sequence_number",
        "_sequence_number_received",
        "_session_status_handler",
        "_user_password",
        "initialized",
        "public_key",
        "session_id",
        "user_id",
    )

    def __init__(
        self,
        remote_addr: tuple[str, int],
        user_id: int,
        user_password: str,
        device_authentication_password: str | None = None,
        connection_lost_cb: Callable[[], None] | None = None,
    ) -> None:
        """Initialize SecureSession class."""
        super().__init__(
            remote_addr=remote_addr,
            connection_lost_cb=connection_lost_cb,
        )
        self._device_authentication_code: bytes | None = (
            derive_device_authentication_password(device_authentication_password)
            if device_authentication_password
            else None
        )
        self.user_id = user_id
        self._user_password = derive_user_password(user_password)

        self._private_key: X25519PrivateKey
        self.public_key: bytes
        self._peer_public_key: X25519PublicKey
        self._key: bytes  # Session Key
        self.session_id: int

        self._sequence_number: int = 0
        self._sequence_number_received: int = -1
        self.initialized: bool = False
        self._keepalive_task: asyncio.Task[None] | None = None
        self._session_status_handler: KNXIPTransport.Callback | None = None

    async def connect(self) -> None:
        """Connect transport."""
        await super().connect()
        self._private_key, self.public_key = generate_ecdh_key_pair()
        self._sequence_number = 0
        self._sequence_number_received = -1
        request_session = Session(
            transport=self,
            ecdh_client_public_key=self.public_key,
        )
        await request_session.start()
        if request_session.response is None:
            raise CommunicationError(
                "Secure session could not be established. No response received."
            )
        # SessionAuthenticate and everything else after now shall be wrapped in SecureWrapper
        authenticate_mac = self.handshake(request_session.response)
        self.initialized = True

        request_authentication = Authenticate(
            transport=self,
            user_id=self.user_id,
            message_authentication_code=authenticate_mac,
        )
        await request_authentication.start()
        if request_authentication.response is None:
            raise CommunicationError(
                "Secure session could not be established. No response received."
            )
        if (  # TODO: look for status in request/response and use `success` instead of response ?
            request_authentication.response.status
            != SecureSessionStatusCode.STATUS_AUTHENTICATION_SUCCESS
        ):
            raise IPSecureError(
                f"Secure session authentication failed: {request_authentication.response.status}"
            )
        self._session_status_handler = self.register_callback(
            self._handle_session_status, [KNXIPServiceType.SESSION_STATUS]
        )

    def handshake(self, session_response: SessionResponse) -> bytes:
        """
        Handshake with device.

        Returns a SessionAuthenticate KNX/IP body.
        """
        self._peer_public_key = X25519PublicKey.from_public_bytes(
            session_response.ecdh_server_public_key
        )
        self.session_id = session_response.secure_session_id
        # verify SessionResponse MAC
        # TODO: get header data from actual KNX/IP frame
        response_header_data = bytes.fromhex("06 10 09 52 00 38")
        pub_keys_xor = bytes_xor(
            self.public_key,
            session_response.ecdh_server_public_key,
        )
        if self._device_authentication_code:
            response_mac_cbc = calculate_message_authentication_code_cbc(
                key=self._device_authentication_code,
                additional_data=response_header_data
                + self.session_id.to_bytes(2, "big")
                + pub_keys_xor,  # knx_ip_header + secure_session_id + bytes_xor(client_pub_key, server_pub_key)
            )
            _, mac_tr = decrypt_ctr(
                key=self._device_authentication_code,
                counter_0=COUNTER_0_HANDSHAKE,
                mac=session_response.message_authentication_code,
            )
            if mac_tr != response_mac_cbc:
                raise IPSecureError("SessionResponse MAC verification failed.")
        # calculate session key
        ecdh_shared_secret = self._private_key.exchange(self._peer_public_key)
        self._key = sha256_hash(ecdh_shared_secret)[:16]
        # generate SessionAuthenticate MAC
        authenticate_header_data = bytes.fromhex("06 10 09 53 00 18")
        authenticate_mac_cbc = calculate_message_authentication_code_cbc(
            key=self._user_password,
            additional_data=authenticate_header_data
            + bytes(1)  # reserved
            + self.user_id.to_bytes(1, "big")
            + pub_keys_xor,
            block_0=bytes(16),
        )
        _, authenticate_mac = encrypt_data_ctr(
            key=self._user_password,
            counter_0=COUNTER_0_HANDSHAKE,
            mac_cbc=authenticate_mac_cbc,
        )
        return authenticate_mac

    def stop(self) -> None:
        """Stop session and transport."""
        if self._session_status_handler:
            self.unregister_callback(self._session_status_handler)
            self._session_status_handler = None
        if self.transport and self.initialized:
            self.send(
                KNXIPFrame.init_from_body(
                    SessionStatus(status=SecureSessionStatusCode.STATUS_CLOSE)
                )
            )
        self.stop_keepalive_task()
        self.initialized = False
        super().stop()

    def handle_knxipframe(self, knxipframe: KNXIPFrame, source: HPAI) -> None:
        """Handle KNXIP Frame and call all callbacks matching the service type ident."""
        # TODO: disallow unencrypted frames with exceptions for discovery etc. eg. DescriptionResponse
        if isinstance(knxipframe.body, SecureWrapper):
            if not self.initialized:
                raise CouldNotParseKNXIP(
                    "Received SecureWrapper while Secure session not initialized"
                )
            new_sequence_number = int.from_bytes(
                knxipframe.body.sequence_information, "big"
            )
            if not new_sequence_number > self._sequence_number_received:
                ip_secure_logger.warning(
                    "Discarding SecureWrapper with invalid sequence number: %s",
                    knxipframe,
                )
                return
            try:
                knxipframe = self.decrypt_frame(knxipframe)
            except KNXSecureValidationError as err:
                ip_secure_logger.warning("Could not decrypt KNXIPFrame: %s", err)
                # Frame shall be discarded
                return
            except CouldNotParseKNXIP as couldnotparseknxip:
                knx_logger.debug(
                    "Unsupported encrypted KNXIPFrame: %s",
                    couldnotparseknxip.description,
                )
                return
            self._sequence_number_received = new_sequence_number
            knx_logger.debug("Decrypted frame: %s", knxipframe)
        super().handle_knxipframe(knxipframe, source)

    async def _session_keepalive(self) -> None:
        """Keep session alive."""
        await asyncio.sleep(SESSION_KEEPALIVE_RATE)
        self.send(
            KNXIPFrame.init_from_body(
                SessionStatus(status=SecureSessionStatusCode.STATUS_KEEPALIVE)
            )
        )

    def send(self, knxipframe: KNXIPFrame, addr: tuple[str, int] | None = None) -> None:
        """Send KNXIPFrame to socket. `addr` is ignored on TCP."""
        if self.initialized:
            knx_logger.debug("Encrypting frame: %s", knxipframe)
            knxipframe = self.encrypt_frame(plain_frame=knxipframe)
            # keepalive timer is started with first and reset with every other
            # SecureWrapper frame (including wrapped keepalive frames themselves)
            self.start_keepalive_task()
        # TODO: disallow sending unencrypted frames over non-initialized session with
        # exceptions for discovery and SessionRequest
        super().send(knxipframe, addr)

    def get_sequence_information(self) -> bytes:
        """Increment sequence number. Return byte representation of current sequence number."""
        next_sn = self._sequence_number.to_bytes(6, "big")
        self._sequence_number += 1
        return next_sn

    def get_message_tag(self) -> bytes:
        """Return byte representation of current message tag."""
        return MESSAGE_TAG_TUNNELLING

    def start_keepalive_task(self) -> None:
        """Start or restart session keepalive task."""
        if self._keepalive_task:
            self._keepalive_task.cancel()
        self._keepalive_task = asyncio.create_task(self._session_keepalive())

    def stop_keepalive_task(self) -> None:
        """Stop keepalive task."""
        if self._keepalive_task:
            self._keepalive_task.cancel()
            self._keepalive_task = None

    def _handle_session_status(
        self, knxipframe: KNXIPFrame, source: HPAI, transport: KNXIPTransport
    ) -> None:
        """Handle session status."""
        assert isinstance(knxipframe.body, SessionStatus)
        if knxipframe.body.status in (
            SecureSessionStatusCode.STATUS_CLOSE,
            SecureSessionStatusCode.STATUS_TIMEOUT,
            SecureSessionStatusCode.STATUS_UNAUTHENTICATED,
        ):
            logger.info("Secure session closed by server: %s.", knxipframe.body.status)
            if self.transport:
                # closing transport will call `asyncio.Protocol.connection_lost`
                # and its callback from SecureTunnel
                self.transport.close()


class SecureGroup(UDPTransport, _IPSecureTransportLayer):
    """Class for secure KNXnet/IP multicast communication."""

    __slots__ = ("_key", "secure_timer")
    session_id = 0  # Routing uses fixed session id 0

    def __init__(
        self,
        local_addr: tuple[str, int],
        remote_addr: tuple[str, int],
        backbone_key: bytes,
        latency_ms: int = 1000,
    ) -> None:
        """Initialize SecureGroup class."""
        super().__init__(
            local_addr=local_addr,
            remote_addr=remote_addr,
            multicast=True,
        )
        self._key = backbone_key
        self.secure_timer = SecureSequenceTimer(
            backbone_key=backbone_key,
            latency_ms=latency_ms,
            transport_send=super().send,
        )

    async def connect(self) -> None:
        """Connect transport."""
        await super().connect()
        await self.secure_timer.synchronize()

    def stop(self) -> None:
        """Stop tasks and transport."""
        self.secure_timer.stop()
        super().stop()

    def handle_knxipframe(self, knxipframe: KNXIPFrame, source: HPAI) -> None:
        """Handle KNXIP Frame and call all callbacks matching the service type ident."""
        if isinstance(knxipframe.body, RoutingIndication):
            ip_secure_logger.info(
                "Discarding received unencrypted RoutingIndication: %s",
                knxipframe,
            )
            return
        if isinstance(knxipframe.body, TimerNotify):
            self.secure_timer.handle_timer_notify(knxipframe.body)
            return
        if isinstance(knxipframe.body, SecureWrapper):
            if not self.secure_timer.timer_authenticated:
                ip_secure_logger.debug(
                    "Discarding received SecureWrapper before timer synchronisazion finished: %s",
                    knxipframe,
                )
                return
            secure_wrapper = knxipframe.body
            try:
                knxipframe = self.decrypt_frame(knxipframe)
            except KNXSecureValidationError as err:
                ip_secure_logger.warning("Could not decrypt KNXIPFrame: %s", err)
                # Frame shall be discarded
                return
            except CouldNotParseKNXIP as couldnotparseknxip:
                knx_logger.debug(
                    "Unsupported encrypted KNXIPFrame: %s",
                    couldnotparseknxip.description,
                )
                return
            if not self.secure_timer.validate_secure_wrapper(secure_wrapper):
                ip_secure_logger.warning(
                    "Discarding SecureWrapper with invalid timer value: %s",
                    secure_wrapper,
                )
                return
            knx_logger.debug("Decrypted frame: %s", knxipframe)
        # handle decrypted frames and plain frames (e.g. SearchRequest for discovery)
        super().handle_knxipframe(knxipframe, source)

    def send(self, knxipframe: KNXIPFrame, addr: tuple[str, int] | None = None) -> None:
        """Send KNXIPFrame to socket. `addr` is ignored on TCP."""
        knx_logger.debug("Encrypting frame: %s", knxipframe)
        knxipframe = self.encrypt_frame(plain_frame=knxipframe)
        # TODO: disallow sending unencrypted frames over non-initialized session with
        # exceptions for discovery and SessionRequest
        super().send(knxipframe, addr)

    def get_sequence_information(self) -> bytes:
        """Return byte representation of current timer value."""
        return self.secure_timer.get_for_outgoing_secure_wrapper().to_bytes(6, "big")

    def get_message_tag(self) -> bytes:
        """Return byte representation of current message tag."""
        return random.randbytes(2)


class SecureSequenceTimer:
    """
    Class for holding and synchronizing the timer for secure sequence information.

    According to AN159 v06 KNXnet-IP Secure AS §2.2.2.3 Timer synchronizing
    """

    __slots__ = (
        "_backbone_key",
        "_clock_difference",
        "_expected_notify_handler",
        "_loop",
        "_notify_timer_handle",
        "_transport_send",
        "latency_tolerance_ms",
        "max_delay_time_follower_periodic_notify",
        "max_delay_time_follower_update_notify",
        "max_delay_time_keeper_periodic_notify",
        "max_delay_time_keeper_update_notify",
        "min_delay_time_follower_periodic_notify",
        "min_delay_time_follower_update_notify",
        "sched_update",
        "sync_latency_tolerance_ms",
        "timekeeper",
        "timer_authenticated",
    )

    TIMER_NOTIFY_HEADER = bytes.fromhex("06 10 09 55 00 24")

    MIN_DELAY_TIME_KEEPER_PERIODIC_NOTIFY: Final = 10  # seconds
    MIN_DELAY_TIME_KEEPER_UPDATE_NOTIFY: Final = 0.1  # seconds
    SYNC_LATENCY_FRACTION: Final = 10  # 10%

    def __init__(
        self,
        backbone_key: bytes,
        latency_ms: int,
        transport_send: Callable[[KNXIPFrame, tuple[str, int] | None], None],
    ) -> None:
        """Initialize SecureSequenceTimer class."""
        self._backbone_key = backbone_key
        self._clock_difference: int = 0
        self._expected_notify_handler: (
            tuple[bytes, asyncio.Future[int]]  # message_tag, synchronization future
            | None
        ) = None
        self._loop = asyncio.get_running_loop()
        self._notify_timer_handle: asyncio.TimerHandle | None = None
        self._transport_send = transport_send

        self.sched_update: bool = False
        self.timekeeper: bool = False
        self.timer_authenticated: bool = False

        self.latency_tolerance_ms = latency_ms
        self.sync_latency_tolerance_ms: int = round(
            latency_ms / 100 * self.SYNC_LATENCY_FRACTION
        )
        _sync_latency_tolerance_seconds = self.sync_latency_tolerance_ms / 1000
        self.max_delay_time_keeper_periodic_notify = (
            self.MIN_DELAY_TIME_KEEPER_PERIODIC_NOTIFY
            + _sync_latency_tolerance_seconds * 3
        )
        self.min_delay_time_follower_periodic_notify = (
            self.max_delay_time_keeper_periodic_notify
            + _sync_latency_tolerance_seconds * 1
        )
        self.max_delay_time_follower_periodic_notify = (
            self.min_delay_time_follower_periodic_notify
            + _sync_latency_tolerance_seconds * 10
        )
        self.max_delay_time_keeper_update_notify = (
            self.MIN_DELAY_TIME_KEEPER_UPDATE_NOTIFY
            + _sync_latency_tolerance_seconds * 1
        )
        self.min_delay_time_follower_update_notify = (
            self.max_delay_time_keeper_update_notify
            + _sync_latency_tolerance_seconds * 1
        )
        self.max_delay_time_follower_update_notify = (
            self.min_delay_time_follower_update_notify
            + _sync_latency_tolerance_seconds * 10
        )

    def stop(self) -> None:
        """Cancel notify timer."""
        if self._notify_timer_handle:
            self._notify_timer_handle.cancel()
            self._notify_timer_handle = None
        if self._expected_notify_handler:
            self._expected_notify_handler[1].cancel()

    def _monotonic_ms(self) -> int:
        """Return current monotonic time in milliseconds."""
        return int(self._loop.time() * 1000.0)

    def current_timer_value(self) -> int:
        """Return current timer value in milliseconds."""
        return self._monotonic_ms() + self._clock_difference

    def update(self, new_value: int) -> None:
        """Update timer value."""
        self._clock_difference = new_value - self._monotonic_ms()

    def reschedule(self, update: tuple[bytes, bytes] | None = None) -> None:
        """
        Reschedule notify timer.

        If `update` is set, the timer is rescheduled for an update notify,
        else for a periodic notify.
        `update` is expected to be a tuple of (message_tag, serial_number).
        """
        if self._notify_timer_handle is not None:
            self._notify_timer_handle.cancel()
        self.sched_update = bool(update)
        min_delay: float
        max_delay: float
        if self.sched_update:
            if self.timekeeper:
                min_delay = self.MIN_DELAY_TIME_KEEPER_UPDATE_NOTIFY
                max_delay = self.max_delay_time_keeper_update_notify
            else:
                min_delay = self.min_delay_time_follower_update_notify
                max_delay = self.max_delay_time_follower_update_notify
        elif self.timekeeper:
            min_delay = self.MIN_DELAY_TIME_KEEPER_PERIODIC_NOTIFY
            max_delay = self.max_delay_time_keeper_periodic_notify
        else:
            min_delay = self.min_delay_time_follower_periodic_notify
            max_delay = self.max_delay_time_follower_periodic_notify
        self._notify_timer_handle = self._loop.call_later(
            random.uniform(min_delay, max_delay), self._notify_timer_expired, update
        )

    def _notify_timer_expired(self, update: tuple[bytes, bytes] | None) -> None:
        """Notify timer expired."""
        if update:
            self.send_timer_notify(message_tag=update[0], serial_number=update[1])
        else:
            self.send_timer_notify()
        if not self.timekeeper:
            self.timekeeper = True
            ip_secure_logger.debug("Becoming timekeeper")
        self.reschedule()

    def send_timer_notify(
        self,
        message_tag: bytes | None = None,
        serial_number: bytes = XKNX_SERIAL_NUMBER,
    ) -> None:
        """Send a TimerNotify frame."""
        timer_value = self.current_timer_value()
        timer_bytes = timer_value.to_bytes(6, "big")
        _message_tag = message_tag or random.randbytes(2)
        b_0 = timer_bytes + serial_number + _message_tag + b"\x00\x00"
        # TODO: get header data and total_length from TimerNotify class
        #       or handle mac calculation in TimerNotify class directly
        mac_cbc = calculate_message_authentication_code_cbc(
            key=self._backbone_key,
            additional_data=self.TIMER_NOTIFY_HEADER,
            block_0=b_0,
        )
        c_0 = timer_bytes + serial_number + _message_tag + b"\xff\x00"
        _, mac = encrypt_data_ctr(
            key=self._backbone_key,
            counter_0=c_0,
            mac_cbc=mac_cbc,
        )
        self._transport_send(
            KNXIPFrame.init_from_body(
                TimerNotify(
                    timer_value=timer_value,
                    serial_number=serial_number,
                    message_tag=_message_tag,
                    message_authentication_code=mac,
                )
            ),
            None,
        )

    def verify_timer_notify_mac(self, timer_notify: TimerNotify) -> None:
        """Verify MAC of timer notify."""
        timer_bytes = timer_notify.timer_value.to_bytes(6, "big")
        b_0 = (
            timer_bytes
            + timer_notify.serial_number
            + timer_notify.message_tag
            + b"\x00\x00"
        )
        c_0 = (
            timer_bytes
            + timer_notify.serial_number
            + timer_notify.message_tag
            + b"\xff\x00"
        )
        _, mac_tr = decrypt_ctr(
            key=self._backbone_key,
            counter_0=c_0,
            mac=timer_notify.message_authentication_code,
        )
        mac_cbc = calculate_message_authentication_code_cbc(
            key=self._backbone_key,
            additional_data=self.TIMER_NOTIFY_HEADER,
            block_0=b_0,
        )
        if mac_cbc != mac_tr:
            raise KNXSecureValidationError("MAC verification failed")

    async def synchronize(self) -> None:
        """Synchronize timer with remote time keeper."""
        message_tag = random.randbytes(2)
        waiter_fut: asyncio.Future[int] = self._loop.create_future()
        self._expected_notify_handler = message_tag, waiter_fut
        self.send_timer_notify(message_tag=message_tag)
        try:
            async with asyncio_timeout(  # 3.3 seconds at latency_ms=1000, SYNC_LATENCY_FRACTION=10%
                self.max_delay_time_follower_update_notify
                + 2 * self.latency_tolerance_ms / 1000
            ):
                timer_value = await waiter_fut
            self.update(new_value=timer_value)
        except asyncio.TimeoutError:
            # use highest received timer value of TimerNotify or SecureWrapper frames
            ip_secure_logger.warning(
                "Timer synchronization not answered. Becoming time keeper."
            )
            self.timekeeper = True
        except asyncio.CancelledError:
            return
        finally:
            self._expected_notify_handler = None
        self.timer_authenticated = True
        self.reschedule()

    def validate_secure_wrapper(self, secure_wrapper: SecureWrapper) -> bool:
        """Validate a SecureWrapper frames and handle timer update schedule."""
        local_timer_value = self.current_timer_value()
        received_timer_value = int.from_bytes(
            secure_wrapper.sequence_information, "big"
        )
        # §2.2.2.3.2.5 Events: E5 - E8
        if received_timer_value > local_timer_value:
            self._clock_difference += received_timer_value - local_timer_value
            if not self.sched_update:
                self.reschedule()
            return True
        if received_timer_value > local_timer_value - self.sync_latency_tolerance_ms:
            if not self.sched_update:
                self.reschedule()
            return True
        if received_timer_value > local_timer_value - self.latency_tolerance_ms:
            return True
        if not self.sched_update:
            self.reschedule(
                update=(secure_wrapper.message_tag, secure_wrapper.serial_number)
            )
        return False

    def handle_timer_notify(self, timer_notify: TimerNotify) -> None:
        """Handle received TimerNotify frame."""
        local_timer_value = self.current_timer_value()
        try:
            self.verify_timer_notify_mac(timer_notify)
        except KNXSecureValidationError:
            ip_secure_logger.warning(
                "Discarding TimerNotify with invalid MAC: %s", timer_notify
            )
            return
        received_timer_value = timer_notify.timer_value
        # §2.2.2.3.2.5 Events: E11
        if (
            self._expected_notify_handler is not None
            and timer_notify.serial_number == XKNX_SERIAL_NUMBER
            and timer_notify.message_tag == self._expected_notify_handler[0]
        ):
            fut = self._expected_notify_handler[1]
            fut.set_result(received_timer_value)
            return
        # §2.2.2.3.2.5 Events: E1 - E4
        if received_timer_value > local_timer_value:
            self._clock_difference += received_timer_value - local_timer_value
            if self.timekeeper:
                ip_secure_logger.debug("Becoming time follower")
                self.timekeeper = False
            self.reschedule()
            return
        if received_timer_value > local_timer_value - self.sync_latency_tolerance_ms:
            if self.timekeeper:
                ip_secure_logger.debug("Becoming time follower")
                self.timekeeper = False
            self.reschedule()
            return
        if received_timer_value > local_timer_value - self.latency_tolerance_ms:
            return
        if not self.sched_update:
            self.reschedule(
                update=(timer_notify.message_tag, timer_notify.serial_number)
            )

    def get_for_outgoing_secure_wrapper(self) -> int:
        """Return current timer value and handle timer update schedule."""
        # §2.2.2.3.2.5 Events: E9
        if not self.sched_update:
            self.reschedule()
        return self.current_timer_value()



"""Abstraction to send SessionAuthenticate and wait for SessionStatus."""

from __future__ import annotations

from typing import TYPE_CHECKING

from xknx.knxip import KNXIPFrame, SessionAuthenticate, SessionStatus

from .request_response import RequestResponse

if TYPE_CHECKING:
    from xknx.io.transport import KNXIPTransport


class Authenticate(RequestResponse):
    """Class to send a SessionAuthenticate and wait for SessionStatus."""

    def __init__(
        self,
        transport: KNXIPTransport,
        user_id: int,
        message_authentication_code: bytes,
    ) -> None:
        """Initialize Session class."""
        super().__init__(transport, SessionStatus, timeout_in_seconds=10)
        self.user_id = user_id
        self.message_authentication_code = message_authentication_code
        self.response: SessionStatus | None = None

    def create_knxipframe(self) -> KNXIPFrame:
        """Create KNX/IP Frame object to be sent to device."""
        return KNXIPFrame.init_from_body(
            SessionAuthenticate(
                user_id=self.user_id,
                message_authentication_code=self.message_authentication_code,
            )
        )

    def on_success_hook(self, knxipframe: KNXIPFrame) -> None:
        """Set communication channel and identifier after having received a valid answer."""
        assert isinstance(knxipframe.body, SessionStatus)
        self.response = knxipframe.body



"""Example on how to disconnect/reset all available tunneling channels."""

import asyncio

from xknx import XKNX
from xknx.io import GatewayScanner
from xknx.io.request_response import ConnectionState, Disconnect
from xknx.io.transport import UDPTransport
from xknx.knxip import HPAI


async def main() -> None:
    """Search for a Tunnelling device, walk through all possible channels and disconnect them."""
    xknx = XKNX()
    gatewayscanner = GatewayScanner(xknx)
    gateways = await gatewayscanner.scan()

    if not gateways:
        print("No Gateways found")
        return

    gateway = gateways[0]

    if not gateway.supports_tunnelling:
        print("Gateway does not support tunneling")
        return

    udp_transport = UDPTransport((gateway.local_ip, 0), (gateway.ip_addr, gateway.port))

    await udp_transport.connect()
    local_hpai = HPAI(*udp_transport.getsockname())

    for i in range(255):
        conn_state = ConnectionState(
            udp_transport, communication_channel_id=i, local_hpai=local_hpai
        )

        await conn_state.start()

        if conn_state.success:
            print("Disconnecting ", i)
            disconnect = Disconnect(
                udp_transport, communication_channel_id=i, local_hpai=local_hpai
            )

            await disconnect.start()

            if disconnect.success:
                print("Disconnected ", i)


asyncio.run(main())



"""Unit test for Switch objects."""

from unittest.mock import AsyncMock, Mock, patch

from xknx import XKNX
from xknx.devices import Device, Sensor
from xknx.dpt import DPTArray
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueRead, GroupValueResponse, GroupValueWrite


@patch.multiple(Device, __abstractmethods__=set())
class TestDevice:
    """Test class for Switch object."""

    def test_device_updated_cb(self) -> None:
        """Test device updated cb is added to the device."""
        xknx = XKNX()
        after_update_callback = Mock()
        device = Device(xknx, "TestDevice", device_updated_cb=after_update_callback)

        device.after_update()
        after_update_callback.assert_called_with(device)

    def test_process_callback(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if callback was called."""
        xknx = XKNX()
        device = Device(xknx, "TestDevice")
        after_update_callback1 = Mock()
        after_update_callback2 = Mock()
        device.register_device_updated_cb(after_update_callback1)
        device.register_device_updated_cb(after_update_callback2)

        # Triggering first time. Both have to be called
        device.after_update()
        after_update_callback1.assert_called_with(device)
        after_update_callback2.assert_called_with(device)
        after_update_callback1.reset_mock()
        after_update_callback2.reset_mock()

        # Triggering 2nd time. Both have to be called
        device.after_update()
        after_update_callback1.assert_called_with(device)
        after_update_callback2.assert_called_with(device)
        after_update_callback1.reset_mock()
        after_update_callback2.reset_mock()

        # Unregistering first callback
        device.unregister_device_updated_cb(after_update_callback1)
        device.after_update()
        after_update_callback1.assert_not_called()
        after_update_callback2.assert_called_with(device)
        after_update_callback1.reset_mock()
        after_update_callback2.reset_mock()

        # Unregistering second callback
        device.unregister_device_updated_cb(after_update_callback2)
        device.after_update()
        after_update_callback1.assert_not_called()
        after_update_callback2.assert_not_called()

    @patch("logging.Logger.exception")
    def test_bad_callback(self, logging_exception_mock: Mock) -> None:
        """Test handling callback raising an exception."""
        xknx = XKNX()
        device = Device(xknx, "TestDevice")
        good_callback_1 = Mock()
        bad_callback = Mock(side_effect=Exception("Boom"))
        good_callback_2 = Mock()

        device.register_device_updated_cb(good_callback_1)
        device.register_device_updated_cb(bad_callback)
        device.register_device_updated_cb(good_callback_2)

        device.after_update()
        good_callback_1.assert_called_with(device)
        bad_callback.assert_called_with(device)
        good_callback_2.assert_called_with(device)

        logging_exception_mock.assert_called_once_with(
            "Unexpected error while processing device_updated_cb for %s",
            device,
        )

    async def test_process(self) -> None:
        """Test if telegram is handled by the correct process_* method."""
        xknx = XKNX()
        device = Device(xknx, "TestDevice")

        with patch("xknx.devices.Device.process_group_read") as mock_group_read:
            telegram = Telegram(
                destination_address=GroupAddress("1/2/1"), payload=GroupValueRead()
            )
            device.process(telegram)
            mock_group_read.assert_called_with(telegram)

        with patch("xknx.devices.Device.process_group_write") as mock_group_write:
            telegram = Telegram(
                destination_address=GroupAddress("1/2/1"),
                payload=GroupValueWrite(DPTArray((0x01, 0x02))),
            )
            device.process(telegram)
            mock_group_write.assert_called_with(telegram)

        with patch("xknx.devices.Device.process_group_response") as mock_group_response:
            telegram = Telegram(
                destination_address=GroupAddress("1/2/1"),
                payload=GroupValueResponse(DPTArray((0x01, 0x02))),
            )
            device.process(telegram)
            mock_group_response.assert_called_with(telegram)

    async def test_sync_with_wait(self) -> None:
        """Test sync with wait_for_result=True."""
        xknx = XKNX()
        sensor = Sensor(
            xknx, "Sensor", group_address_state="1/2/3", value_type="wind_speed_ms"
        )

        with patch(
            "xknx.remote_value.RemoteValue.read_state", new_callable=AsyncMock
        ) as read_state_mock:
            await sensor.sync(wait_for_result=True)
            read_state_mock.assert_called_with(wait_for_result=True)

    async def test_process_group_write(self) -> None:
        """Test if process_group_write. Nothing really to test here."""
        xknx = XKNX()
        device = Device(xknx, "TestDevice")
        device.process_group_write(Telegram(destination_address=GroupAddress(1)))

    async def test_process_group_response(self) -> None:
        """Test if process_group_read. Testing if mapped to group_write."""
        xknx = XKNX()
        device = Device(xknx, "TestDevice")
        with patch("xknx.devices.Device.process_group_write") as mock_group_write:
            device.process_group_response(Telegram(destination_address=GroupAddress(1)))
            mock_group_write.assert_called_with(
                Telegram(destination_address=GroupAddress(1))
            )

    async def test_process_group_read(self) -> None:
        """Test if process_group_read. Nothing really to test here."""
        xknx = XKNX()
        device = Device(xknx, "TestDevice")
        device.process_group_read(Telegram(destination_address=GroupAddress(1)))



"""Unit test for Address class."""

from typing import Any

import pytest

from xknx.exceptions import CouldNotParseAddress
from xknx.telegram.address import (
    GroupAddress,
    GroupAddressType,
    IndividualAddress,
    InternalGroupAddress,
    parse_device_group_address,
)

_broadcast_group_addresses: list[str | int] = ["0/0/0", "0/0", "0", 0]

device_group_addresses_valid: dict[Any, int] = {
    "0/1": 1,
    "0/11": 11,
    "0/111": 111,
    "0/1111": 1111,
    "0/2047": 2047,
    "0/0/1": 1,
    "0/0/11": 11,
    "0/0/111": 111,
    "0/0/255": 255,
    "0/1/11": 267,
    "0/1/111": 367,
    "0/7/255": 2047,
    "1/0": 2048,
    "1/0/0": 2048,
    "1/1/111": 2415,
    "1/7/255": 4095,
    "31/7/255": 65535,
    "1": 1,
    65535: 65535,
    GroupAddress("1/1/111"): 2415,
}

group_addresses_valid = (
    dict.fromkeys(_broadcast_group_addresses, 0) | device_group_addresses_valid
)

group_addresses_invalid: list[Any] = [
    "0/2049",
    "0/8/0",
    "0/0/256",
    "32/0",
    "0/0a",
    "a0/0",
    "abc",
    "1.1.1",
    "0.0",
    IndividualAddress("11.11.111"),
    65536,
    (0xAB, 0xCD),
    -1,
    [],
    None,
]

device_group_addresses_invalid: list[Any] = [
    *group_addresses_invalid,
    *_broadcast_group_addresses,
]

individual_addresses_valid: dict[Any, int] = {
    "0.0.0": 0,
    "123": 123,
    "1.0.0": 4096,
    "1.1.0": 4352,
    "1.1.1": 4353,
    "1.1.11": 4363,
    "1.1.111": 4463,
    "1.11.111": 7023,
    "11.11.111": 47983,
    IndividualAddress("11.11.111"): 47983,
    "15.15.255": 65535,
    0: 0,
    65535: 65535,
}

individual_addresses_invalid: list[Any] = [
    "15.15.256",
    "16.0.0",
    "0.16.0",
    "15.15.255a",
    "a15.15.255",
    "abc",
    "1/1/1",
    "0/0",
    GroupAddress("1/1/111"),
    65536,
    (0xAB, 0xCD),
    -1,
    [],
    None,
]

internal_group_addresses_valid: dict[str | InternalGroupAddress, str] = {
    "i 123": "i-123",
    "i-123": "i-123",
    "i_123": "i-123",
    "I 123": "i-123",
    "i abc": "i-abc",
    "i-abc": "i-abc",
    "i_abc": "i-abc",
    "I-abc": "i-abc",
    "i123": "i-123",
    "iabc": "i-abc",
    "IABC": "i-ABC",
    "i   abc  ": "i-abc",
    "i asdf 123 adsf ": "i-asdf 123 adsf",
    "i-1/2/3": "i-1/2/3",
    InternalGroupAddress("i-123"): "i-123",
}

internal_group_addresses_invalid: list[str | None] = [
    "i",
    "i-",
    "i ",
    "i  ",
    "i- ",
    "a",
    None,
]


class TestIndividualAddress:
    """Test class for IndividualAddress."""

    @pytest.mark.parametrize(
        ("address_test", "address_raw"), individual_addresses_valid.items()
    )
    def test_with_valid(self, address_test: Any, address_raw: int) -> None:
        """Test with some valid addresses."""

        assert IndividualAddress(address_test).raw == address_raw

    @pytest.mark.parametrize("address_test", individual_addresses_invalid)
    def test_with_invalid(self, address_test: Any) -> None:
        """Test with some invalid addresses."""

        with pytest.raises(CouldNotParseAddress):
            IndividualAddress(address_test)

    def test_with_int(self) -> None:
        """Test initialization with free format address as integer."""
        assert IndividualAddress(49552).raw == 49552

    def test_from_knx(self) -> None:
        """Test initialization with Bytes."""
        ia = IndividualAddress.from_knx(bytes((0x12, 0x34)))
        assert ia.raw == 0x1234

    def test_is_line(self) -> None:
        """Test if `IndividualAddress.is_line` works like excepted."""
        assert IndividualAddress("1.0.0").is_line
        assert not IndividualAddress("1.0.1").is_line

    def test_is_device(self) -> None:
        """Test if `IndividualAddress.is_device` works like excepted."""
        assert IndividualAddress("1.0.1").is_device
        assert not IndividualAddress("1.0.0").is_device

    def test_to_knx(self) -> None:
        """Test if `IndividualAddress.to_knx()` generates valid byte objects."""
        assert IndividualAddress("0.0.0").to_knx() == bytes((0x0, 0x0))
        assert IndividualAddress("15.15.255").to_knx() == bytes((0xFF, 0xFF))

    def test_equal(self) -> None:
        """Test if the equal operator works in all cases."""
        assert IndividualAddress("1.0.0") == IndividualAddress(4096)
        assert IndividualAddress("1.0.0") != IndividualAddress("1.1.1")
        assert IndividualAddress("1.0.0") is not None
        assert IndividualAddress("1.0.0") != "example"
        assert IndividualAddress("1.1.1") != GroupAddress("1/1/1")
        assert IndividualAddress(250) != GroupAddress(250)
        assert IndividualAddress(250) != 250

    def test_comparison(self) -> None:
        """Test if the less than operator works in all cases."""
        assert IndividualAddress("1.0.0") < IndividualAddress("1.0.1")
        assert IndividualAddress("1.1.1") < IndividualAddress("1.10.0")
        assert IndividualAddress("1.1.0") < IndividualAddress("2.0.0")
        with pytest.raises(
            TypeError,
            match=r"'<' not supported between instances of 'IndividualAddress' and 'GroupAddress'",
        ):
            _ = IndividualAddress("1.0.0") < GroupAddress("1/0")

    def test_representation(self) -> None:
        """Test string representation of address."""
        assert repr(IndividualAddress("2.3.4")) == 'IndividualAddress("2.3.4")'


class TestGroupAddress:
    """Test class for GroupAddress."""

    @pytest.mark.parametrize(
        ("address_test", "address_raw"), group_addresses_valid.items()
    )
    def test_with_valid(self, address_test: str | int, address_raw: int) -> None:
        """
        Test if the class constructor generates valid raw values.

        This test checks:
        * all allowed input variants (strings, tuples, integers)
        * for conversation errors
        * for upper/lower limits still working, to avoid off-by-one errors
        """

        assert GroupAddress(address_test).raw == address_raw

    @pytest.mark.parametrize("address_test", group_addresses_invalid)
    def test_with_invalid(self, address_test: Any) -> None:
        """
        Test if constructor raises an exception for all known invalid cases.

        Checks:
        * addresses or parts of it too high/low
        * invalid input variants (lists instead of tuples)
        * invalid strings
        """

        with pytest.raises(CouldNotParseAddress):
            GroupAddress(address_test)

    def test_main(self) -> None:
        """
        Test if `GroupAddress.main` works.

        Checks:
        * Main group part of a strings returns the right value
        * Return `None` on `GroupAddressType.FREE`
        """
        assert GroupAddress("1/0").main == 1
        assert GroupAddress("15/0").main == 15
        assert GroupAddress("31/0/0").main == 31
        GroupAddress.address_format = GroupAddressType.FREE
        assert GroupAddress("1/0").main is None

    def test_middle(self) -> None:
        """
        Test if `GroupAddress.middle` works.

        Checks:
        * Middle group part of a strings returns the right value
        * Return `None` if not `GroupAddressType.LONG`
        """
        GroupAddress.address_format = GroupAddressType.LONG
        assert GroupAddress("1/0/1").middle == 0
        assert GroupAddress("1/7/1").middle == 7
        GroupAddress.address_format = GroupAddressType.SHORT
        assert GroupAddress("1/0").middle is None
        GroupAddress.address_format = GroupAddressType.FREE
        assert GroupAddress("1/0").middle is None

    def test_sub(self) -> None:
        """
        Test if `GroupAddress.sub` works.

        Checks:
        * Sub group part of a strings returns the right value
        * Return never `None`
        """
        GroupAddress.address_format = GroupAddressType.SHORT
        assert GroupAddress("1/0").sub == 0
        assert GroupAddress("31/0").sub == 0
        assert GroupAddress("1/2047").sub == 2047
        assert GroupAddress("31/2047").sub == 2047

        GroupAddress.address_format = GroupAddressType.LONG
        assert GroupAddress("1/0/0").sub == 0
        assert GroupAddress("1/0/255").sub == 255

        GroupAddress.address_format = GroupAddressType.FREE
        assert GroupAddress("0/0").sub == 0
        assert GroupAddress("1/0").sub == 2048
        assert GroupAddress("31/2047").sub == 65535

    def test_from_knx(self) -> None:
        """Test initialization with Bytes."""
        ga = GroupAddress.from_knx(bytes((0x12, 0x34)))
        assert ga.raw == 0x1234

    def test_to_knx(self) -> None:
        """Test if `GroupAddress.to_knx()` generates valid byte tuples."""
        assert GroupAddress("0/0").to_knx() == bytes((0x0, 0x0))
        assert GroupAddress("31/2047").to_knx() == bytes((0xFF, 0xFF))

    def test_equal(self) -> None:
        """Test if the equal operator works in all cases."""
        assert GroupAddress("1/0") == GroupAddress(2048)
        assert GroupAddress("1/1") != GroupAddress("1/1/0")
        assert GroupAddress("1/0") is not None
        assert GroupAddress("1/0") != "example"
        assert GroupAddress(1) != IndividualAddress(1)
        assert GroupAddress(1) != 1

    def test_comparison(self) -> None:
        """Test if the less than operator works in all cases."""
        assert GroupAddress("1/0/4") < GroupAddress("1/1/0")
        assert GroupAddress("1/1/1") < GroupAddress("10/0/0")
        assert GroupAddress("1/2047") < GroupAddress("2/0")
        assert GroupAddress("1/0/1") >= GroupAddress("1/0/1")
        with pytest.raises(
            TypeError,
            match=r"'<' not supported between instances of 'GroupAddress' and 'IndividualAddress'",
        ):
            _ = GroupAddress("1/0") < IndividualAddress("1.0.0")
        with pytest.raises(
            TypeError,
            match=r"'<' not supported between instances of 'GroupAddress' and 'int'",
        ):
            _ = GroupAddress("1/0") < 100
        with pytest.raises(
            TypeError,
            match=r"'<' not supported between instances of 'GroupAddress' and 'InternalGroupAddress'",
        ):
            _ = GroupAddress("1/0") < InternalGroupAddress("i-0")

        assert sorted(
            [GroupAddress("1/0/1"), GroupAddress("1/0/0"), GroupAddress("0/20")]
        ) == [GroupAddress("0/20"), GroupAddress("1/0/0"), GroupAddress("1/0/1")]

    @pytest.mark.parametrize(
        ("initializer", "string_free", "string_short", "string_long"),
        [
            ("0", "0", "0/0", "0/0/0"),
            ("1/4/70", "3142", "1/1094", "1/4/70"),
        ],
    )
    def test_representation(
        self, initializer: str, string_free: str, string_short: str, string_long: str
    ) -> None:
        """Test string representation of address."""
        group_address = GroupAddress(initializer)
        GroupAddress.address_format = GroupAddressType.FREE
        assert str(group_address) == string_free
        assert repr(group_address) == f'GroupAddress("{string_free}")'
        GroupAddress.address_format = GroupAddressType.SHORT
        assert str(group_address) == string_short
        assert repr(group_address) == f'GroupAddress("{string_short}")'
        GroupAddress.address_format = GroupAddressType.LONG
        assert str(group_address) == string_long
        assert repr(group_address) == f'GroupAddress("{string_long}")'


class TestInternalGroupAddress:
    """Test class for InternalGroupAddress."""

    @pytest.mark.parametrize(
        ("address_test", "address_raw"), internal_group_addresses_valid.items()
    )
    def test_with_valid(
        self, address_test: str | InternalGroupAddress, address_raw: str
    ) -> None:
        """Test if the class constructor generates valid raw values."""

        assert InternalGroupAddress(address_test).raw == address_raw

    @pytest.mark.parametrize(
        "address_test",
        [
            *internal_group_addresses_invalid,
            *group_addresses_valid,
            *group_addresses_invalid,
            *individual_addresses_valid,
            *individual_addresses_invalid,
        ],
    )
    def test_with_invalid(self, address_test: Any) -> None:
        """Test if constructor raises an exception for all known invalid cases."""

        with pytest.raises(CouldNotParseAddress):
            InternalGroupAddress(address_test)

    def test_equal(self) -> None:
        """Test if the equal operator works in all cases."""
        assert InternalGroupAddress("i 123") == InternalGroupAddress("i 123")
        assert InternalGroupAddress("i-asdf") == InternalGroupAddress("i asdf")
        assert InternalGroupAddress("i-asdf") == InternalGroupAddress("Iasdf")
        assert InternalGroupAddress("i-1") != InternalGroupAddress("i-2")
        assert InternalGroupAddress("i-1") is not None
        assert InternalGroupAddress("i-example") != "example"
        assert InternalGroupAddress("i-0") != GroupAddress(0)
        assert InternalGroupAddress("i-1") != IndividualAddress(1)
        assert InternalGroupAddress("i-1") != 1

    def test_comparison(self) -> None:
        """Test if the less than operator works in all cases."""
        assert InternalGroupAddress("i-1") < InternalGroupAddress("i-2")
        assert InternalGroupAddress("i-abc") < InternalGroupAddress("i-def")
        assert InternalGroupAddress("i-1") < InternalGroupAddress("i-a")
        assert InternalGroupAddress("i-7") >= InternalGroupAddress("i-7")
        with pytest.raises(
            TypeError,
            match=r"'<' not supported between instances of 'InternalGroupAddress' and 'GroupAddress'",
        ):
            _ = InternalGroupAddress("i-1") < GroupAddress("1/0")
        with pytest.raises(
            TypeError,
            match=r"'<' not supported between instances of 'InternalGroupAddress' and 'IndividualAddress'",
        ):
            _ = InternalGroupAddress("i-1") < IndividualAddress("1.0.0")
        with pytest.raises(
            TypeError,
            match=r"'<' not supported between instances of 'InternalGroupAddress' and 'str'",
        ):
            _ = InternalGroupAddress("i-1") < "i-2"

    def test_representation(self) -> None:
        """Test string representation of address."""
        assert repr(InternalGroupAddress("i0")) == 'InternalGroupAddress("i-0")'
        assert repr(InternalGroupAddress("i-0")) == 'InternalGroupAddress("i-0")'
        assert repr(InternalGroupAddress("i 0")) == 'InternalGroupAddress("i-0")'


class TestParseDestinationAddress:
    """Test class for parsing destination addresses."""

    @pytest.mark.parametrize("address_test", device_group_addresses_valid)
    def test_parse_device_group_address(self, address_test: Any) -> None:
        """Test if the function returns GroupAddress objects."""
        assert isinstance(parse_device_group_address(address_test), GroupAddress)

    @pytest.mark.parametrize("address_test", internal_group_addresses_valid)
    def test_parse_device_internal_group_address(
        self, address_test: str | InternalGroupAddress
    ) -> None:
        """Test if the function returns InternalGroupAddress objects."""
        assert isinstance(
            parse_device_group_address(address_test), InternalGroupAddress
        )

    @pytest.mark.parametrize(
        "address_test",
        [
            *internal_group_addresses_invalid,
            *device_group_addresses_invalid,
        ],
    )
    def test_parse_device_invalid(self, address_test: Any) -> None:
        """Test if the function raises CouldNotParseAddress on invalid values."""
        with pytest.raises(CouldNotParseAddress):
            parse_device_group_address(address_test)

    def test_parse_device_invalid_group_address_message(self) -> None:
        """Test if the error message is from GroupAddress, not InternalGroupAddress for strings."""
        with pytest.raises(CouldNotParseAddress, match=r".*Sub group out of range.*"):
            parse_device_group_address("0/0/700")



"""Example for sending generic telegrams to the KNX bus."""

import asyncio
import logging

from xknx import XKNX
from xknx.tools import group_value_response, group_value_write

logging.basicConfig(level=logging.INFO)
logging.getLogger("xknx.log").level = logging.DEBUG
logging.getLogger("xknx.tools").level = logging.DEBUG


async def main() -> None:
    """Connect to KNX bus and send telegrams."""
    async with XKNX() as xknx:
        # send a DPT 9.001 temperature value
        group_value_write(xknx, "5/1/20", 21.7, value_type="temperature")
        # send a response DPT 1 binary value
        group_value_response(xknx, "5/1/20", True)


asyncio.run(main())



"""Unit test for TravelCalculator objects."""

from unittest.mock import patch

from xknx.devices import TravelCalculator, TravelStatus


class TestTravelCalculator:
    """Test class for TravelCalculator objects."""

    # TravelCalculator(25, 50) means:
    #
    #     2 steps / sec UP
    #     4 steps / sec DOWN
    #
    # INIT
    #
    def test_init(self) -> None:
        """Test initial status."""
        travelcalculator = TravelCalculator(25, 50)
        assert not travelcalculator.is_closed()
        assert not travelcalculator.is_closing()
        assert not travelcalculator.is_opening()
        assert not travelcalculator.is_traveling()
        assert travelcalculator.position_reached()
        assert travelcalculator.current_position() is None

    def test_set_position(self) -> None:
        """Test set position."""
        travelcalculator = TravelCalculator(25, 50)
        travelcalculator.set_position(70)
        assert travelcalculator.position_reached()
        assert travelcalculator.current_position() == 70

    def test_set_position_after_travel(self) -> None:
        """Set explicit position after start_travel should stop traveling."""
        travelcalculator = TravelCalculator(25, 50)
        travelcalculator.start_travel(30)
        travelcalculator.set_position(80)
        assert travelcalculator.position_reached()
        assert travelcalculator.current_position() == 80

    def test_travel_down(self) -> None:
        """Test travel down."""
        travelcalculator = TravelCalculator(25, 50)
        with patch("time.time") as mock_time:
            mock_time.return_value = 1580000000.0
            travelcalculator.set_position(40)
            travelcalculator.start_travel(60)

            # time not changed, still at beginning
            assert travelcalculator.current_position() == 40
            assert not travelcalculator.position_reached()
            assert travelcalculator.travel_direction == TravelStatus.DIRECTION_DOWN

            mock_time.return_value = 1580000001.0
            assert travelcalculator.current_position() == 44
            assert not travelcalculator.position_reached()

            mock_time.return_value = 1580000002.0
            assert travelcalculator.current_position() == 48
            assert not travelcalculator.position_reached()

            mock_time.return_value = 1580000004.0
            assert travelcalculator.current_position() == 56
            assert not travelcalculator.position_reached()

            mock_time.return_value = 1580000005.0
            # position reached
            assert travelcalculator.current_position() == 60
            assert travelcalculator.position_reached()

            mock_time.return_value = 1580000010.0
            assert travelcalculator.current_position() == 60
            assert travelcalculator.position_reached()

    def test_travel_up(self) -> None:
        """Test travel up."""
        travelcalculator = TravelCalculator(25, 50)
        with patch("time.time") as mock_time:
            mock_time.return_value = 1580000000.0
            travelcalculator.set_position(70)
            travelcalculator.start_travel(50)

            # time not changed, still at beginning
            assert travelcalculator.current_position() == 70
            assert not travelcalculator.position_reached()
            assert travelcalculator.travel_direction == TravelStatus.DIRECTION_UP

            mock_time.return_value = 1580000002.0
            assert travelcalculator.current_position() == 66
            assert not travelcalculator.position_reached()

            mock_time.return_value = 1580000004.0
            assert travelcalculator.current_position() == 62
            assert not travelcalculator.position_reached()

            mock_time.return_value = 1580000008.0
            assert travelcalculator.current_position() == 54
            assert not travelcalculator.position_reached()

            mock_time.return_value = 1580000010.0
            # position reached
            assert travelcalculator.current_position() == 50
            assert travelcalculator.position_reached()

            mock_time.return_value = 1580000020.0
            assert travelcalculator.current_position() == 50
            assert travelcalculator.position_reached()

    def test_stop(self) -> None:
        """Test stopping."""
        travelcalculator = TravelCalculator(25, 50)
        with patch("time.time") as mock_time:
            mock_time.return_value = 1580000000.0
            travelcalculator.set_position(80)
            travelcalculator.start_travel(60)
            assert travelcalculator.travel_direction == TravelStatus.DIRECTION_UP

            # stop aftert two seconds
            mock_time.return_value = 1580000002.0
            travelcalculator.stop()

            mock_time.return_value = 1580000004.0
            assert travelcalculator.current_position() == 76
            assert travelcalculator.position_reached()

            # restart after 1 additional second (3 seconds)
            mock_time.return_value = 1580000005.0
            travelcalculator.start_travel(68)

            # running up for 6 seconds
            mock_time.return_value = 1580000006.0
            assert travelcalculator.current_position() == 74
            assert not travelcalculator.position_reached()

            mock_time.return_value = 1580000009.0
            assert travelcalculator.current_position() == 68
            assert travelcalculator.position_reached()

    def test_travel_down_with_updates(self) -> None:
        """Test travel down with position updates from bus."""
        travelcalculator = TravelCalculator(25, 50)
        with patch("time.time") as mock_time:
            mock_time.return_value = 1580000000.0
            travelcalculator.set_position(40)
            travelcalculator.start_travel(100)  # 15 seconds to reach 100

            # time not changed, still at beginning
            assert travelcalculator.current_position() == 40
            assert not travelcalculator.position_reached()
            assert travelcalculator.travel_direction == TravelStatus.DIRECTION_DOWN

            mock_time.return_value = 1580000002.0
            assert travelcalculator.current_position() == 48
            assert not travelcalculator.position_reached()
            # update from bus matching calculation
            travelcalculator.update_position(48)
            assert travelcalculator.current_position() == 48
            assert not travelcalculator.position_reached()

            mock_time.return_value = 1580000010.0
            assert travelcalculator.current_position() == 80
            assert not travelcalculator.position_reached()
            # update from bus not matching calculation takes precedence (1 second slower)
            travelcalculator.update_position(76)
            assert travelcalculator.current_position() == 76
            assert not travelcalculator.position_reached()
            # travel time extended by 1 second due to update from bus
            mock_time.return_value = 1580000015.0
            assert travelcalculator.current_position() == 96
            assert not travelcalculator.position_reached()
            mock_time.return_value = 1580000015.0 + 1
            assert travelcalculator.current_position() == 100
            assert travelcalculator.position_reached()

    def test_travel_up_with_updates(self) -> None:
        """Test travel up with position updates from bus."""
        travelcalculator = TravelCalculator(25, 50)
        with patch("time.time") as mock_time:
            mock_time.return_value = 1580000000.0
            travelcalculator.set_position(70)
            travelcalculator.start_travel(50)  # 10 seconds to reach 50

            mock_time.return_value = 1580000005.0
            assert travelcalculator.current_position() == 60
            assert not travelcalculator.position_reached()
            # update from bus not matching calculation takes precedence (1 second faster)
            travelcalculator.update_position(58)
            assert travelcalculator.current_position() == 58
            assert not travelcalculator.position_reached()
            # position reached 1 second earlier than predicted
            mock_time.return_value = 1580000010.0 - 1
            assert travelcalculator.current_position() == 50
            assert travelcalculator.position_reached()

    def test_change_direction(self) -> None:
        """Test changing direction while travelling."""
        travelcalculator = TravelCalculator(50, 25)
        with patch("time.time") as mock_time:
            mock_time.return_value = 1580000000.0
            travelcalculator.set_position(60)
            travelcalculator.start_travel(80)
            assert travelcalculator.travel_direction == TravelStatus.DIRECTION_DOWN

            # change direction after two seconds
            mock_time.return_value = 1580000002.0
            assert travelcalculator.current_position() == 64
            travelcalculator.start_travel(48)
            assert travelcalculator.travel_direction == TravelStatus.DIRECTION_UP

            assert travelcalculator.current_position() == 64
            assert not travelcalculator.position_reached()

            mock_time.return_value = 1580000004.0
            assert travelcalculator.current_position() == 56
            assert not travelcalculator.position_reached()

            mock_time.return_value = 1580000006.0
            assert travelcalculator.current_position() == 48
            assert travelcalculator.position_reached()

    def test_travel_full_up(self) -> None:
        """Test travelling to the full up position."""
        travelcalculator = TravelCalculator(25, 50)
        with patch("time.time") as mock_time:
            mock_time.return_value = 1580000000.0
            travelcalculator.set_position(30)
            travelcalculator.start_travel_up()

            mock_time.return_value = 1580000014.0
            assert not travelcalculator.position_reached()
            assert not travelcalculator.is_closed()
            assert not travelcalculator.is_open()

            mock_time.return_value = 1580000015.0
            assert travelcalculator.position_reached()
            assert travelcalculator.is_open()
            assert not travelcalculator.is_closed()

    def test_travel_full_down(self) -> None:
        """Test travelling to the full down position."""
        travelcalculator = TravelCalculator(25, 50)
        with patch("time.time") as mock_time:
            mock_time.return_value = 1580000000.0
            travelcalculator.set_position(20)
            travelcalculator.start_travel_down()

            mock_time.return_value = 1580000019.0
            assert not travelcalculator.position_reached()
            assert not travelcalculator.is_closed()
            assert not travelcalculator.is_open()

            mock_time.return_value = 1580000020.0
            assert travelcalculator.position_reached()
            assert travelcalculator.is_closed()
            assert not travelcalculator.is_open()

    def test_is_traveling(self) -> None:
        """Test if cover is traveling and position reached."""
        travelcalculator = TravelCalculator(25, 50)
        with patch("time.time") as mock_time:
            mock_time.return_value = 1580000000.0
            assert not travelcalculator.is_traveling()
            assert travelcalculator.position_reached()

            travelcalculator.set_position(80)
            assert not travelcalculator.is_traveling()
            assert travelcalculator.position_reached()

            mock_time.return_value = 1580000000.0
            travelcalculator.start_travel_down()

            mock_time.return_value = 1580000004.0
            assert travelcalculator.is_traveling()
            assert not travelcalculator.position_reached()

            mock_time.return_value = 1580000005.0
            assert not travelcalculator.is_traveling()
            assert travelcalculator.position_reached()

    def test_is_opening_closing(self) -> None:
        """Test reports is_opening and is_closing."""
        travelcalculator = TravelCalculator(25, 50)
        with patch("time.time") as mock_time:
            mock_time.return_value = 1580000000.0
            assert not travelcalculator.is_opening()
            assert not travelcalculator.is_closing()

            travelcalculator.set_position(80)
            assert not travelcalculator.is_opening()
            assert not travelcalculator.is_closing()

            mock_time.return_value = 1580000000.0
            travelcalculator.start_travel_down()
            assert not travelcalculator.is_opening()
            assert travelcalculator.is_closing()

            mock_time.return_value = 1580000004.0
            assert not travelcalculator.is_opening()
            assert travelcalculator.is_closing()

            mock_time.return_value = 1580000005.0
            assert not travelcalculator.is_opening()
            assert not travelcalculator.is_closing()
            # up direction
            travelcalculator.start_travel(50)
            assert travelcalculator.is_opening()
            assert not travelcalculator.is_closing()

            mock_time.return_value = 1580000030.0
            assert not travelcalculator.is_opening()
            assert not travelcalculator.is_closing()



"""Tests for secure util primitives."""

import pytest

from xknx.secure.util import byte_pad, bytes_xor, sha256_hash


@pytest.mark.parametrize(
    ("value_a", "value_b", "result"),
    [
        (
            "01010101",
            "10101010",
            "11111111",
        ),
        (
            "000011111111111111110000",
            "000000000000111111110000",
            "000011111111000000000000",
        ),
        (
            "000000000000000000000000000000001",
            "000011111111111111110000101010101",
            "000011111111111111110000101010100",
        ),
    ],
)
def test_byte_xor(
    value_a: tuple[str, str, str],
    value_b: tuple[str, str, str],
    result: tuple[str, str, str],
) -> None:
    """Test byte xor."""
    len_a = (len(value_a) + 7) // 8
    len_b = (len(value_b) + 7) // 8
    len_res = max(len_a, len_b)
    assert bytes_xor(
        int(value_a, 2).to_bytes(len_a, "big"),
        int(value_b, 2).to_bytes(len_b, "big"),
    ) == int(result, 2).to_bytes(len_res, "big")


def test_byte_xor_error() -> None:
    """Test byte xor error."""
    with pytest.raises(ValueError):
        bytes_xor(bytes([1]), bytes([0, 1]))


@pytest.mark.parametrize(
    ("block_size", "data", "result"),
    [
        (4, bytes([23]), bytes([23, 0, 0, 0])),
        (4, bytes([1, 23, 0, 0]), bytes([1, 23, 0, 0])),
        (16, bytes([123]), bytes([123, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])),
        (16, bytes(16), bytes(16)),
        (16, bytes(17), bytes(32)),
        (16, bytes(32), bytes(32)),
        (16, bytes(47), bytes(48)),
    ],
)
def test_byte_pad(block_size: int, data: bytes, result: bytes) -> None:
    """Test byte pad."""
    assert byte_pad(data=data, block_size=block_size) == result


def test_sha256_hash() -> None:
    """Test sha256 hash."""
    # Data from SessionResponse example in KNX specification AN159v06
    assert sha256_hash(
        bytes.fromhex(
            "d8 01 52 52 17 61 8f 0d a9 0a 4f f2 21 48 ae e0"
            "ff 4c 19 b4 30 e8 08 12 23 ff e9 9c 81 a9 8b 05"
        )
    ) == bytes.fromhex(
        "28 94 26 c2 91 25 35 ba 98 27 9a 4d 18 43 c4 87"
        "7f 6d 2d c3 7e 40 dc 4b eb fe 40 31 d4 73 3b 30"
    )



"""Unit test for Switch objects."""

from unittest.mock import Mock

from xknx import XKNX
from xknx.devices import Switch
from xknx.dpt import DPTBinary
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueRead, GroupValueResponse, GroupValueWrite

from ..conftest import EventLoopClockAdvancer


class TestSwitch:
    """Test class for Switch object."""

    #
    # SYNC
    #
    async def test_sync(self) -> None:
        """Test sync function / sending group reads to KNX bus."""
        xknx = XKNX()
        switch = Switch(
            xknx, "TestOutlet", group_address_state="1/2/3", group_address="1/2/4"
        )
        await switch.sync()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"), payload=GroupValueRead()
        )

    async def test_sync_state_address(self) -> None:
        """Test sync function / sending group reads to KNX bus. Test with Switch with explicit state address."""
        xknx = XKNX()
        switch = Switch(
            xknx, "TestOutlet", group_address="1/2/3", group_address_state="1/2/4"
        )
        await switch.sync()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/4"), payload=GroupValueRead()
        )

    #
    # TEST PROCESS
    #
    async def test_process(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if device was updated."""
        xknx = XKNX()
        callback_mock = Mock()

        switch1 = Switch(
            xknx, "TestOutlet", group_address="1/2/3", device_updated_cb=callback_mock
        )
        switch2 = Switch(
            xknx, "TestOutlet", group_address="1/2/3", device_updated_cb=callback_mock
        )
        assert switch1.state is None
        assert switch2.state is None
        callback_mock.assert_not_called()

        telegram_on = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        telegram_off = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(0)),
        )

        switch1.process(telegram_on)
        assert switch1.state is True
        callback_mock.assert_called_once()
        callback_mock.reset_mock()
        switch1.process(telegram_off)
        assert switch1.state is False
        callback_mock.assert_called_once()
        callback_mock.reset_mock()
        # test setting switch2 to False with first telegram
        switch2.process(telegram_off)
        assert switch2.state is False
        callback_mock.assert_called_once()
        callback_mock.reset_mock()
        switch2.process(telegram_on)
        assert switch2.state is True
        callback_mock.assert_called_once()
        callback_mock.reset_mock()

    async def test_process_state(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if device was updated."""
        xknx = XKNX()
        callback_mock = Mock()

        switch1 = Switch(
            xknx,
            "TestOutlet",
            group_address="1/2/3",
            group_address_state="1/2/4",
            device_updated_cb=callback_mock,
        )
        switch2 = Switch(
            xknx,
            "TestOutlet",
            group_address="1/2/3",
            group_address_state="1/2/4",
            device_updated_cb=callback_mock,
        )
        assert switch1.state is None
        assert switch2.state is None
        callback_mock.assert_not_called()

        telegram_on = Telegram(
            destination_address=GroupAddress("1/2/4"),
            payload=GroupValueResponse(DPTBinary(1)),
        )
        telegram_off = Telegram(
            destination_address=GroupAddress("1/2/4"),
            payload=GroupValueResponse(DPTBinary(0)),
        )

        switch1.process(telegram_on)
        assert switch1.state is True
        callback_mock.assert_called_once()
        callback_mock.reset_mock()
        switch1.process(telegram_off)
        assert switch1.state is False
        callback_mock.assert_called_once()
        callback_mock.reset_mock()
        # test setting switch2 to False with first telegram
        switch2.process(telegram_off)
        assert switch2.state is False
        callback_mock.assert_called_once()
        callback_mock.reset_mock()
        switch2.process(telegram_on)
        assert switch2.state is True
        callback_mock.assert_called_once()
        callback_mock.reset_mock()

    async def test_process_invert(self) -> None:
        """Test process / reading telegrams from telegram queue with inverted switch."""
        xknx = XKNX()
        switch = Switch(xknx, "TestOutlet", group_address="1/2/3", invert=True)
        assert switch.state is None

        telegram_inv_on = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(0)),
        )
        telegram_inv_off = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )

        switch.process(telegram_inv_on)
        assert switch.state is True
        switch.process(telegram_inv_off)
        assert switch.state is False

    async def test_process_reset_after(
        self, time_travel: EventLoopClockAdvancer
    ) -> None:
        """Test process reset_after."""
        xknx = XKNX()
        reset_after_sec = 1
        switch = Switch(
            xknx, "TestInput", group_address="1/2/3", reset_after=reset_after_sec
        )
        telegram_on = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )

        switch.process(telegram_on)
        assert switch.state
        assert xknx.telegrams.qsize() == 0
        await time_travel(reset_after_sec)
        assert xknx.telegrams.qsize() == 1
        switch.process(xknx.telegrams.get_nowait())
        assert not switch.state

    async def test_process_reset_after_cancel_existing(
        self, time_travel: EventLoopClockAdvancer
    ) -> None:
        """Test process reset_after cancels existing reset tasks."""
        xknx = XKNX()
        reset_after_sec = 0.01
        switch = Switch(
            xknx, "TestInput", group_address="1/2/3", reset_after=reset_after_sec
        )
        telegram_on = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueResponse(DPTBinary(1)),
        )

        switch.process(telegram_on)
        assert switch.state
        assert xknx.telegrams.qsize() == 0
        await time_travel(reset_after_sec / 2)
        # half way through the reset timer
        switch.process(telegram_on)
        assert switch.state

        await time_travel(reset_after_sec / 2)
        assert xknx.telegrams.qsize() == 0

    async def test_remove_device(self, xknx_no_interface: XKNX) -> None:
        """Test device removal cancels task."""
        xknx = xknx_no_interface
        switch = Switch(xknx, "TestInput", group_address="1/2/3", reset_after=1)
        xknx.devices.async_add(switch)
        async with xknx:
            telegram_on = Telegram(
                destination_address=GroupAddress("1/2/3"),
                payload=GroupValueResponse(DPTBinary(1)),
            )
            switch.process(telegram_on)
            assert switch.state
            assert not switch._reset_task.done()
            xknx.devices.async_remove(switch)
            assert switch._reset_task.done()

    async def test_process_callback(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if callback was called."""

        xknx = XKNX()
        switch = Switch(xknx, "TestOutlet", group_address="1/2/3")

        after_update_callback = Mock()
        switch.register_device_updated_cb(after_update_callback)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        switch.process(telegram)

        after_update_callback.assert_called_with(switch)

    #
    # TEST RESPOND
    #
    async def test_respond_to_read(self) -> None:
        """Test respond_to_read function."""
        xknx = XKNX()
        responding = Switch(
            xknx,
            "TestSensor1",
            group_address="1/1/1",
            respond_to_read=True,
        )
        non_responding = Switch(
            xknx,
            "TestSensor2",
            group_address="1/1/1",
            respond_to_read=False,
        )
        responding_multiple = Switch(
            xknx,
            "TestSensor3",
            group_address=["1/1/1", "3/3/3"],
            group_address_state="2/2/2",
            respond_to_read=True,
        )
        #  set initial payload of Switch
        responding.switch.value = True
        non_responding.switch.value = True
        responding_multiple.switch.value = True

        read_telegram = Telegram(
            destination_address=GroupAddress("1/1/1"), payload=GroupValueRead()
        )
        # verify no response when respond is False
        non_responding.process(read_telegram)
        assert xknx.telegrams.qsize() == 0

        # verify response when respond is True
        responding.process(read_telegram)
        assert xknx.telegrams.qsize() == 1
        response = xknx.telegrams.get_nowait()
        assert response == Telegram(
            destination_address=GroupAddress("1/1/1"),
            payload=GroupValueResponse(DPTBinary(True)),
        )
        # verify no response when GroupValueRead request is not for group_address
        responding_multiple.process(read_telegram)
        assert xknx.telegrams.qsize() == 1
        response = xknx.telegrams.get_nowait()
        assert response == Telegram(
            destination_address=GroupAddress("1/1/1"),
            payload=GroupValueResponse(DPTBinary(True)),
        )
        responding_multiple.process(
            Telegram(
                destination_address=GroupAddress("2/2/2"), payload=GroupValueRead()
            )
        )
        responding_multiple.process(
            Telegram(
                destination_address=GroupAddress("3/3/3"), payload=GroupValueRead()
            )
        )
        assert xknx.telegrams.qsize() == 0

    #
    # TEST SET ON
    #
    async def test_set_on(self) -> None:
        """Test switching on switch."""
        xknx = XKNX()
        switch = Switch(xknx, "TestOutlet", group_address="1/2/3")
        await switch.set_on()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )

    #
    # TEST SET OFF
    #
    async def test_set_off(self) -> None:
        """Test switching off switch."""
        xknx = XKNX()
        switch = Switch(xknx, "TestOutlet", group_address="1/2/3")
        await switch.set_off()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(0)),
        )

    #
    # TEST SET INVERT
    #
    async def test_set_invert(self) -> None:
        """Test switching on/off inverted switch."""
        xknx = XKNX()
        switch = Switch(xknx, "TestOutlet", group_address="1/2/3", invert=True)

        await switch.set_on()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(0)),
        )

        await switch.set_off()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )

    #
    # TEST has_group_address
    #
    def test_has_group_address(self) -> None:
        """Test has_group_address."""
        xknx = XKNX()
        switch = Switch(xknx, "TestOutlet", group_address="1/2/3")
        assert switch.has_group_address(GroupAddress("1/2/3"))
        assert not switch.has_group_address(GroupAddress("2/2/2"))

        assert switch.group_addresses() == {GroupAddress("1/2/3")}

    #
    # TEST passive group addresses
    #
    def test_has_group_address_passive(self) -> None:
        """Test has_group_address with passive group address."""
        xknx = XKNX()
        switch = Switch(xknx, "TestOutlet", group_address=["1/2/3", "4/4/4"])
        assert switch.has_group_address(GroupAddress("1/2/3"))
        assert switch.has_group_address(GroupAddress("4/4/4"))
        assert not switch.has_group_address(GroupAddress("2/2/2"))

        assert switch.group_addresses() == {
            GroupAddress("1/2/3"),
            GroupAddress("4/4/4"),
        }

    async def test_process_passive(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if device was updated."""
        xknx = XKNX()
        callback_mock = Mock()

        switch1 = Switch(
            xknx,
            "TestOutlet",
            group_address=["1/2/3", "4/4/4"],
            group_address_state=["1/2/30", "5/5/5"],
            device_updated_cb=callback_mock,
        )
        assert switch1.state is None
        callback_mock.assert_not_called()

        telegram_on_passive = Telegram(
            destination_address=GroupAddress("4/4/4"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        telegram_off_passive = Telegram(
            destination_address=GroupAddress("5/5/5"),
            payload=GroupValueWrite(DPTBinary(0)),
        )

        switch1.process(telegram_on_passive)
        assert switch1.state is True
        callback_mock.assert_called_once()
        callback_mock.reset_mock()
        switch1.process(telegram_off_passive)
        assert switch1.state is False
        callback_mock.assert_called_once()
        callback_mock.reset_mock()



#!/usr/bin/env python
"""
Example of a daemon listening for values from my main power-meter and resend them on a MQTT bus.

This example will not be able to run as it is - but it will hopefully give you some
ideas to how you can define DPT, and get their converted values, and even send them to MQTT
in a tested topic-format.

I have a Mosquitto MQTT Server - and running the Paho Python client.
I use some external MQTT libraries as well to handle the MQTT Topic-creation.

The data published on the MQTT bus can be fetched and stored in InfluxDB for graphing,
or monitored by other listeners - that triggers different events.

Please join XKNX on Discord (https://discord.gg/EuAQDXU) and chat with JohanElmis for
specific questions.
"""

import asyncio
import re
import sys

from xknx.devices import Device

try:
    # The following library is not included.
    from myhouse_sensors_mqtt import MetricType, SensorClientMqtt

    # import time
    import paho.mqtt.client as mqtt

    from xknx import XKNX
    from xknx.devices import Sensor

    # import pprint
except ImportError as import_err:
    err_list = str(import_err).split(" ")
    print(f"Unable to import module: {err_list[3]}")
    print(f"Please install the {err_list[3]} module for Python.")
    sys.exit()

BROKER_ADDRESS = "127.0.0.1"
# Give this client a name on the MQTT bus.
mqttc = mqtt.Client("main_power_central")

# Library to deal with verification of values.
# It also triggers and send values to the MQTT bus if the values has changed.
# With no changes it can go up to max_interval_seconds before it's sent.
# This allows me to fetch fast changes without storing all data at a 15s interval.
mh_sensor = SensorClientMqtt(
    change_trigger_percent=5,
    max_interval_seconds=600,
    metric_class="sensor",
    debug=True,
)

# Pre-compile some regexp filters that will be used to catch the different types.
RE_METER_READING = re.compile("MeterReading_")
RE_ACTIVE_POWER = re.compile("ActivePower")
RE_REACTIVE_POWER = re.compile("ReactivePower")
RE_APPARENT_POWER = re.compile("ApparentPower")
RE_VOLTAGE = re.compile("Voltage_")
RE_CURRENT = re.compile("Current_")
RE_FREQUENCY = re.compile("Frequency_")


def device_updated_cb(device: Device) -> None:
    """Do something with the updated device."""
    # print(device.name + ' ' + str(device.resolve_state()) + ' ' + device.unit_of_measurement())
    value = None
    topic = None
    if re.search("^EL-T-O_", device.name):
        metric = str(device.name)[7:]
        value = device.resolve_state()
        if RE_ACTIVE_POWER.search(metric):
            topic = mh_sensor.get_mqtt_sensor_metric(
                MetricType.WATT, "main_power_central", metric
            )
        elif RE_REACTIVE_POWER.search(metric):
            topic = mh_sensor.get_mqtt_sensor_metric(
                MetricType.VAR, "main_power_central", metric
            )
        elif RE_APPARENT_POWER.search(metric):
            topic = mh_sensor.get_mqtt_sensor_metric(
                MetricType.VA, "main_power_central", metric
            )
        elif RE_VOLTAGE.search(metric):
            topic = mh_sensor.get_mqtt_sensor_metric(
                MetricType.VOLTAGE, "main_power_central", metric
            )
        elif RE_CURRENT.search(metric):
            topic = mh_sensor.get_mqtt_sensor_metric(
                MetricType.CURRENT, "main_power_central", metric
            )
        elif RE_METER_READING.search(metric):
            topic = mh_sensor.get_mqtt_sensor_metric(
                MetricType.KWH, "main_power_central", metric
            )
        elif RE_FREQUENCY:
            topic = mh_sensor.get_mqtt_sensor_metric(
                MetricType.CUSTOM, "main_power_central", metric
            )
    else:
        print(
            f"Uncatched metric: {device.name} {device.resolve_state()!s} {device.unit_of_measurement()}"
        )

    if topic and value:
        # This will create a topic like:
        # myhouse/sensor/KWH/main_power_central/MeterReading_ActiveEnergy 103150280
        # myhouse/sensor/WATT/main_power_central/ActivePower_L2 1028.1099853515625
        # myhouse/sensor/VAR/main_power_central/ReactivePower_L3 136.3699951171875
        # myhouse/sensor/VA/main_power_central/ApparentPower_L3 1449.919921875
        # myhouse/sensor/CURRENT/main_power_central/Current_L3 6.239999294281006

        # When storing it to InfluxDB I have a DB named 'myhouse' where all these values will be stored.
        # Then I use the next two fields as metric name (lowercase):  sensor/current
        # break out main_power_central as a tag 'location', and the last field is tagged as the
        # 'sensor': ApparentPower_L3

        # This makes it really easy to graph in for example Grafana.
        # My latest version of the library doesn't send the value after the MQTT Topic, but a JSON structure
        # that also contains time.

        print(f"{topic} {value!s}")
        # ts = int(time.time() * 1000)
        mqttc.publish(topic, value)


async def main() -> None:
    """
    KNX device objects are created and the MQTT server connection is established.

    Then the XKNX Daemon will be started.
    Then everything else happens in the device_updated-function above as it is triggered when we receive data.
    """
    # Connect to KNX/IP device and listen if a switch was updated via KNX bus.
    xknx = XKNX(device_updated_cb=device_updated_cb, daemon_mode=True)

    # The KNX addresses to monitor are defined below, but is normally placed in an external
    #  file that is loaded in on start.

    # Generic Types not specifically supported by XKNX
    Sensor(
        xknx,
        "EL-T-O_MeterReading_ActiveEnergy",
        group_address_state="5/6/11",
        value_type="DPT-13",
    )
    Sensor(
        xknx,
        "EL-T-O_MeterReading_ReactiveEnergy",
        group_address_state="5/6/16",
        value_type="DPT-13",
    )

    # Active Power
    Sensor(
        xknx,
        "EL-T-O_TotalActivePower",
        group_address_state="5/6/24",
        value_type="power",
    )
    Sensor(
        xknx, "EL-T-O_ActivePower_L1", group_address_state="5/6/25", value_type="power"
    )
    # ...

    # Reactive Power
    Sensor(
        xknx,
        "EL-T-O_TotalReactivePower",
        group_address_state="5/6/28",
        value_type="power",
    )
    Sensor(
        xknx,
        "EL-T-O_ReactivePower_L1",
        group_address_state="5/6/29",
        value_type="power",
    )
    # ...

    # Apparent Power
    Sensor(
        xknx,
        "EL-T-O_TotalReactivePower",
        group_address_state="5/6/32",
        value_type="power",
    )
    Sensor(
        xknx,
        "EL-T-O_ApparentPower_L1",
        group_address_state="5/6/33",
        value_type="power",
    )
    # ...

    # Current
    Sensor(
        xknx,
        "EL-T-O_Current_L1",
        group_address_state="5/6/45",
        value_type="electric_current",
    )
    # ...

    # Voltage
    Sensor(
        xknx,
        "EL-T-O_Voltage_L1-N",
        group_address_state="5/6/48",
        value_type="electric_potential",
    )
    # ...

    # Frequency
    Sensor(
        xknx, "EL-T-O_Frequency", group_address_state="5/6/53", value_type="frequency"
    )

    mqttc.connect(BROKER_ADDRESS, 8883, 60)
    mqttc.loop_start()

    # Wait until Ctrl-C was pressed
    await xknx.start()

    await xknx.stop()
    await mqttc.loop_stop()
    await mqttc.disconnect()


asyncio.run(main())



"""Unit test for Cover objects."""

from unittest.mock import AsyncMock, Mock, patch

from xknx import XKNX
from xknx.devices import Cover
from xknx.dpt import DPTArray, DPTBinary
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueRead, GroupValueWrite

from ..conftest import EventLoopClockAdvancer


class TestCover:
    """Test class for Cover objects."""

    #
    # SUPPORTS STOP/POSITION/ANGLE
    #
    def test_supports_stop_true(self) -> None:
        """Test support_position_true."""
        xknx = XKNX()
        cover_short_stop = Cover(
            xknx,
            "Children.Venetian",
            group_address_long="1/4/14",
            group_address_short="1/4/15",
        )
        assert cover_short_stop.supports_stop

        cover_manual_stop = Cover(
            xknx,
            "Children.Venetian",
            group_address_long="1/4/14",
            group_address_stop="1/4/15",
        )
        assert cover_manual_stop.supports_stop

    async def test_supports_stop_false(self) -> None:
        """Test support_position_true."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "Children.Venetian",
            group_address_long="1/4/14",
            group_address_position="1/4/16",
            group_address_angle="1/4/18",
        )
        assert not cover.supports_stop
        with patch("logging.Logger.warning") as mock_warn:
            await cover.stop()
            assert xknx.telegrams.qsize() == 0
            mock_warn.assert_called_with(
                "Stop not supported for device %s", "Children.Venetian"
            )

    def test_supports_position_true(self) -> None:
        """Test support_position_true."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "Children.Venetian",
            group_address_long="1/4/14",
            group_address_short="1/4/15",
            group_address_position="1/4/16",
        )
        assert cover.supports_position

    def test_supports_position_false(self) -> None:
        """Test support_position_true."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "Children.Venetian",
            group_address_long="1/4/14",
            group_address_short="1/4/15",
        )
        assert not cover.supports_position

    def test_supports_angle_true(self) -> None:
        """Test support_position_true."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "Children.Venetian",
            group_address_long="1/4/14",
            group_address_short="1/4/15",
            group_address_angle="1/4/18",
        )
        assert cover.supports_angle

    def test_support_angle_false(self) -> None:
        """Test support_position_true."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "Children.Venetian",
            group_address_long="1/4/14",
            group_address_short="1/4/15",
        )
        assert not cover.supports_angle

    #
    # SUPPORTS LOCKED
    #
    def test_support_locked(self) -> None:
        """Test support_position_true."""
        xknx = XKNX()
        cover_locked = Cover(
            xknx,
            "Children.Venetian",
            group_address_locked_state="1/4/14",
        )
        assert cover_locked.supports_locked
        cover_manual_stop = Cover(
            xknx,
            "Children.Venetian",
            group_address_long="1/4/14",
            group_address_stop="1/4/15",
        )
        assert cover_manual_stop.supports_locked is False

    #
    # SYNC
    #
    async def test_sync(self) -> None:
        """Test sync function / sending group reads to KNX bus."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverSync",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_position_state="1/2/3",
        )
        await cover.sync()
        assert xknx.telegrams.qsize() == 1
        telegram1 = xknx.telegrams.get_nowait()
        assert telegram1 == Telegram(
            destination_address=GroupAddress("1/2/3"), payload=GroupValueRead()
        )

    async def test_sync_state(self) -> None:
        """Test sync function with explicit state address."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverSyncState",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_position="1/2/3",
            group_address_position_state="1/2/4",
        )
        await cover.sync()
        assert xknx.telegrams.qsize() == 1
        telegram1 = xknx.telegrams.get_nowait()
        assert telegram1 == Telegram(
            destination_address=GroupAddress("1/2/4"), payload=GroupValueRead()
        )

    async def test_sync_angle(self) -> None:
        """Test sync function for cover with angle."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverSyncAngle",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_position_state="1/2/3",
            group_address_angle_state="1/2/4",
        )
        await cover.sync()
        assert xknx.telegrams.qsize() == 2
        telegram1 = xknx.telegrams.get_nowait()
        assert telegram1 == Telegram(
            destination_address=GroupAddress("1/2/3"), payload=GroupValueRead()
        )
        telegram2 = xknx.telegrams.get_nowait()
        assert telegram2 == Telegram(
            destination_address=GroupAddress("1/2/4"), payload=GroupValueRead()
        )

    async def test_sync_angle_state(self) -> None:
        """Test sync function with angle/explicit state."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverSyncAngleState",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_angle="1/2/3",
            group_address_angle_state="1/2/4",
        )
        await cover.sync()
        assert xknx.telegrams.qsize() == 1
        telegram1 = xknx.telegrams.get_nowait()
        assert telegram1 == Telegram(
            destination_address=GroupAddress("1/2/4"), payload=GroupValueRead()
        )

    #
    # TEST SET UP
    #
    async def test_set_up(self) -> None:
        """Test moving cover to 'up' position."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverSetUp",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_position="1/2/3",
            group_address_position_state="1/2/4",
        )
        await cover.set_up()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        # DPT 1.008 - 0:up 1:down
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/1"),
            payload=GroupValueWrite(DPTBinary(0)),
        )

    #
    # TEST SET DOWN
    #
    async def test_set_short_down(self) -> None:
        """Test moving cover to 'down' position."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverShortDown",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_position="1/2/3",
            group_address_position_state="1/2/4",
        )
        await cover.set_down()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/1"),
            payload=GroupValueWrite(DPTBinary(1)),
        )

    #
    # TEST SET DOWN INVERTED
    #
    async def test_set_down_inverted(self) -> None:
        """Test moving cover to 'down' position."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverSetDownInverted",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_position="1/2/3",
            group_address_position_state="1/2/4",
            invert_updown=True,
        )
        await cover.set_down()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/1"),
            payload=GroupValueWrite(DPTBinary(0)),
        )

    #
    # TEST SET SHORT UP
    #
    async def test_set_short_up(self) -> None:
        """Test moving cover 'short up'."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverSetShortUp",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_position="1/2/3",
            group_address_position_state="1/2/4",
        )
        await cover.set_short_up()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        # DPT 1.008 - 0:up 1:down
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/2"),
            payload=GroupValueWrite(DPTBinary(0)),
        )

    #
    # TEST SET UP INVERTED
    #
    async def test_set_up_inverted(self) -> None:
        """Test moving cover 'short up'."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverSetUpInverted",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_position="1/2/3",
            group_address_position_state="1/2/4",
            invert_updown=True,
        )
        await cover.set_short_up()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        # DPT 1.008 - 0:up 1:down
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/2"),
            payload=GroupValueWrite(DPTBinary(1)),
        )

    #
    # TEST SET SHORT DOWN
    #
    async def test_set_down(self) -> None:
        """Test moving cover 'short down'."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverSetDown",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_position="1/2/3",
            group_address_position_state="1/2/4",
        )
        await cover.set_short_down()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        # DPT 1.008 - 0:up 1:down
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/2"),
            payload=GroupValueWrite(DPTBinary(1)),
        )

    #
    # TEST STOP
    #
    async def test_stop(self) -> None:
        """Test stopping cover."""
        xknx = XKNX()
        cover_short_stop = Cover(
            xknx,
            "TestCoverStop",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_position="1/2/3",
            group_address_position_state="1/2/4",
        )
        # Attempt stopping while not actually moving
        await cover_short_stop.stop()
        assert xknx.telegrams.qsize() == 0

        # Attempt stopping while moving down
        cover_short_stop.travelcalculator.set_position(0)
        await cover_short_stop.set_down()
        await cover_short_stop.stop()
        assert xknx.telegrams.qsize() == 2
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/1"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/2"),
            payload=GroupValueWrite(DPTBinary(1)),
        )

        # Attempt stopping while moving up
        cover_short_stop.travelcalculator.set_position(100)
        await cover_short_stop.set_up()
        await cover_short_stop.stop()
        assert xknx.telegrams.qsize() == 2
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/1"),
            payload=GroupValueWrite(DPTBinary(0)),
        )
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/2"),
            payload=GroupValueWrite(DPTBinary(0)),
        )

        cover_manual_stop = Cover(
            xknx,
            "TestCoverManualStop",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_stop="1/2/0",
            group_address_position="1/2/3",
            group_address_position_state="1/2/4",
        )
        await cover_manual_stop.stop()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/0"),
            payload=GroupValueWrite(DPTBinary(1)),
        )

    async def test_stop_angle(self) -> None:
        """Test stopping cover during angle move / tilting."""
        xknx = XKNX()
        cover_short_stop = Cover(
            xknx,
            "TestCoverStopAngle",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_angle="1/2/5",
            group_address_angle_state="1/2/6",
        )
        # Attempt stopping while not actually tilting
        await cover_short_stop.stop()
        assert xknx.telegrams.qsize() == 0

        # Set cover tilt to a dummy start value, since otherwise we cannot
        # determine later on a tilt direction and without it, stopping the
        # til process has no effect.
        cover_short_stop.angle.process(
            Telegram(
                destination_address=GroupAddress("1/2/5"),
                payload=GroupValueWrite(DPTArray(0xAA)),
            )
        )

        # Attempt stopping while tilting down
        await cover_short_stop.set_angle(100)
        await cover_short_stop.stop()
        assert xknx.telegrams.qsize() == 2
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTArray(0xFF)),
        )
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/2"),
            payload=GroupValueWrite(DPTBinary(1)),
        )

        # Attempt stopping while tilting up
        await cover_short_stop.set_angle(0)
        await cover_short_stop.stop()
        assert xknx.telegrams.qsize() == 2
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTArray(0x00)),
        )
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/2"),
            payload=GroupValueWrite(DPTBinary(0)),
        )

    #
    # TEST POSITION
    #
    async def test_position(self) -> None:
        """Test moving cover to absolute position."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverPosition",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_position="1/2/3",
            group_address_position_state="1/2/4",
        )
        await cover.set_position(50)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray(0x80)),
        )
        await cover.stop()  # clean up tasks

    async def test_position_without_binary(self) -> None:
        """Test moving cover - with no binary positioning supported."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverPositionWithoutBinary",
            group_address_position="1/2/3",
        )
        await cover.set_down()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray(0xFF)),
        )
        await cover.set_position(50)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray(0x80)),
        )
        await cover.set_up()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray(0x00)),
        )

    async def test_position_without_position_address_up(self) -> None:
        """Test moving cover to absolute position - with no absolute positioning supported."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverPWPAD",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_position_state="1/2/4",
        )
        cover.travelcalculator.set_position(60)
        await cover.set_position(50)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        # DPT 1.008 - 0:up 1:down
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/1"),
            payload=GroupValueWrite(DPTBinary(0)),
        )
        assert cover.travelcalculator._travel_to_position == 50
        assert cover.is_opening()
        # process the outgoing telegram to make sure it doesn't overwrite the target position
        cover.process(telegram)
        assert cover.travelcalculator._travel_to_position == 50
        assert xknx.telegrams.qsize() == 0

        await cover.stop()  # clean up tasks

    async def test_position_without_position_address_down(self) -> None:
        """Test moving cover down - with no absolute positioning supported."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverPWPAD",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_position_state="1/2/4",
        )
        cover.travelcalculator.set_position(70)
        await cover.set_position(80)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/1"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        assert cover.travelcalculator._travel_to_position == 80
        assert cover.is_closing()
        # process the outgoing telegram to make sure it doesn't overwrite the target position
        cover.process(telegram)
        assert cover.travelcalculator._travel_to_position == 80

        await cover.stop()  # clean up tasks

    async def test_position_without_position_address_uninitialized_up(self) -> None:
        """Test moving uninitialized cover to absolute position - with no absolute positioning supported."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverPWPAUU",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_position_state="1/2/4",
        )

        with patch("logging.Logger.warning") as mock_warn:
            await cover.set_position(50)
            assert xknx.telegrams.qsize() == 0
            mock_warn.assert_called_with(
                "Current position unknown. Initialize cover by moving to end position."
            )

        await cover.set_position(0)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/1"),
            payload=GroupValueWrite(DPTBinary(0)),
        )
        await cover.stop()  # clean up tasks

    async def test_position_without_position_address_uninitialized_down(self) -> None:
        """Test moving uninitialized cover to absolute position - with no absolute positioning supported."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverPWPAUD",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_position_state="1/2/4",
        )

        with patch("logging.Logger.warning") as mock_warn:
            await cover.set_position(50)
            assert xknx.telegrams.qsize() == 0
            mock_warn.assert_called_with(
                "Current position unknown. Initialize cover by moving to end position."
            )

        await cover.set_position(100)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/1"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        await cover.stop()  # clean up tasks

    async def test_angle(self) -> None:
        """Test changing angle."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "Children.Venetian",
            group_address_long="1/4/14",
            group_address_short="1/4/15",
            group_address_position_state="1/4/17",
            group_address_position="1/4/16",
            group_address_angle="1/4/18",
            group_address_angle_state="1/4/19",
        )
        await cover.set_angle(50)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/4/18"),
            payload=GroupValueWrite(DPTArray(0x80)),
        )

    async def test_angle_not_supported(self) -> None:
        """Test changing angle on cover which does not support angle."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "Children.Venetian",
            group_address_long="1/4/14",
            group_address_short="1/4/15",
        )
        with patch("logging.Logger.warning") as mock_warn:
            await cover.set_angle(50)
            assert xknx.telegrams.qsize() == 0
            mock_warn.assert_called_with(
                "Angle not supported for device %s", "Children.Venetian"
            )

    #
    # TEST PROCESS
    #
    async def test_process_position(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if position is processed correctly."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverProcessPosition",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_position="1/2/3",
            group_address_position_state="1/2/4",
        )
        # initial position process - position is unknown so this is the new state - not moving
        telegram = Telegram(
            GroupAddress("1/2/3"), payload=GroupValueWrite(DPTArray(213))
        )
        cover.process(telegram)
        assert cover.current_position() == 84
        assert not cover.is_traveling()
        # state telegram updates current position - we are not moving so this is new state - not moving
        telegram = Telegram(
            GroupAddress("1/2/4"), payload=GroupValueWrite(DPTArray(42))
        )
        cover.process(telegram)
        assert cover.current_position() == 16
        assert not cover.is_traveling()
        assert cover.travelcalculator._travel_to_position == 16
        # new position - movement starts
        telegram = Telegram(
            GroupAddress("1/2/3"), payload=GroupValueWrite(DPTArray(255))
        )
        cover.process(telegram)
        assert cover.current_position() == 16
        assert cover.is_closing()
        assert cover.travelcalculator._travel_to_position == 100
        # new state while moving - movement goes on; travelcalculator updated
        telegram = Telegram(
            GroupAddress("1/2/4"), payload=GroupValueWrite(DPTArray(213))
        )
        cover.process(telegram)
        assert cover.current_position() == 84
        assert cover.is_closing()
        assert cover.travelcalculator._travel_to_position == 100

        await cover.stop()  # clean up tasks

    async def test_process_angle(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if position is processed correctly."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverProcessAngle",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_angle="1/2/3",
            group_address_angle_state="1/2/4",
        )
        telegram = Telegram(
            GroupAddress("1/2/4"), payload=GroupValueWrite(DPTArray(42))
        )
        cover.process(telegram)
        assert cover.current_angle() == 16

    async def test_process_locked(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if position is processed correctly."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverProcessLocked",
            group_address_long="1/2/1",
            group_address_locked_state="1/2/4",
        )
        telegram = Telegram(
            GroupAddress("1/2/4"), payload=GroupValueWrite(DPTBinary(1))
        )
        cover.process(telegram)
        assert cover.is_locked() is True

    async def test_process_up(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if up/down is processed correctly."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverProcessUp",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
        )
        cover.travelcalculator.set_position(50)
        assert not cover.is_traveling()
        telegram = Telegram(
            GroupAddress("1/2/1"), payload=GroupValueWrite(DPTBinary(0))
        )
        cover.process(telegram)
        assert cover.is_opening()

        await cover.stop()  # clean up tasks

    async def test_process_down(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if up/down is processed correctly."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverProcessDown",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
        )
        cover.travelcalculator.set_position(50)
        assert not cover.is_traveling()
        telegram = Telegram(
            GroupAddress("1/2/1"), payload=GroupValueWrite(DPTBinary(1))
        )
        cover.process(telegram)
        assert cover.is_closing()

        await cover.stop()  # clean up tasks

    async def test_process_stop(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if stop is processed correctly."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverProcessStop",
            group_address_long="1/2/1",
            group_address_stop="1/2/2",
        )
        cover.travelcalculator.set_position(50)
        await cover.set_down()
        assert cover.is_traveling()
        telegram = Telegram(
            GroupAddress("1/2/2"), payload=GroupValueWrite(DPTBinary(1))
        )
        cover.process(telegram)
        assert not cover.is_traveling()

    async def test_process_short_stop(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if stop is processed correctly."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverProcessShortStop",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
        )
        cover.travelcalculator.set_position(50)
        await cover.set_down()
        assert cover.is_traveling()
        telegram = Telegram(
            GroupAddress("1/2/2"), payload=GroupValueWrite(DPTBinary(1))
        )
        cover.process(telegram)
        assert not cover.is_traveling()

    async def test_process_callback(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if callback is executed."""

        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverProcessCallback",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_stop="1/2/3",
            group_address_position="1/2/4",
            group_address_position_state="1/2/5",
            group_address_angle="1/2/6",
            group_address_angle_state="1/2/7",
        )

        after_update_callback = Mock()

        cover.register_device_updated_cb(after_update_callback)
        for address, payload, _feature in [
            ("1/2/1", DPTBinary(1), "long"),
            ("1/2/2", DPTBinary(1), "short"),
            ("1/2/4", DPTArray(42), "position"),
            ("1/2/3", DPTBinary(1), "stop"),
            # call position with same value again to make sure `always_callback` is set for target position
            ("1/2/4", DPTArray(42), "position"),
            ("1/2/5", DPTArray(42), "position state"),
            ("1/2/6", DPTArray(42), "angle"),
            ("1/2/7", DPTArray(51), "angle state"),
        ]:
            telegram = Telegram(
                destination_address=GroupAddress(address),
                payload=GroupValueWrite(payload),
            )
            cover.process(telegram)
            after_update_callback.assert_called_with(cover)
            after_update_callback.reset_mock()
        # Stop only when cover is travelling
        telegram = Telegram(
            GroupAddress("1/2/3"), payload=GroupValueWrite(DPTBinary(1))
        )
        cover.process(telegram)
        after_update_callback.assert_not_called()
        await cover.set_down()
        cover.process(telegram)
        after_update_callback.assert_called_with(cover)

        await cover.stop()  # clean up tasks

    #
    # IS TRAVELING / IS UP / IS DOWN
    #
    async def test_is_traveling(self) -> None:
        """Test moving cover to absolute position."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverIsTraveling",
            group_address_long="1/2/1",
            group_address_stop="1/2/2",
            group_address_position="1/2/3",
            group_address_position_state="1/2/4",
            travel_time_down=10,
            travel_time_up=10,
        )
        with patch("time.time") as mock_time:
            mock_time.return_value = 1517000000.0
            assert not cover.is_traveling()
            assert not cover.is_opening()
            assert not cover.is_closing()
            assert cover.position_reached()
            # we start with state open covers (up)
            cover.travelcalculator.set_position(0)
            await cover.set_down()
            assert cover.is_traveling()
            assert cover.is_open()
            assert not cover.is_closed()
            assert not cover.is_opening()
            assert cover.is_closing()

            mock_time.return_value = 1517000005.0  # 5 Seconds, half way
            assert not cover.position_reached()
            assert cover.is_traveling()
            assert not cover.is_open()
            assert not cover.is_closed()
            assert not cover.is_opening()
            assert cover.is_closing()

            mock_time.return_value = 1517000010.0  # 10 Seconds, fully closed
            assert cover.position_reached()
            assert not cover.is_traveling()
            assert not cover.is_open()
            assert cover.is_closed()
            assert not cover.is_opening()
            assert not cover.is_closing()
            # up again
            await cover.set_up()
            assert not cover.position_reached()
            assert cover.is_traveling()
            assert not cover.is_open()
            assert cover.is_closed()
            assert cover.is_opening()
            assert not cover.is_closing()

            mock_time.return_value = 1517000015.0  # 15 Seconds, half way
            assert not cover.position_reached()
            assert cover.is_traveling()
            assert not cover.is_open()
            assert not cover.is_closed()
            assert cover.is_opening()
            assert not cover.is_closing()

            mock_time.return_value = 1517000016.0  # 16 Seconds, manual stop
            await cover.stop()
            assert cover.position_reached()
            assert not cover.is_traveling()
            assert not cover.is_open()
            assert not cover.is_closed()
            assert not cover.is_opening()
            assert not cover.is_closing()

    #
    # TEST TASKS
    #
    async def test_auto_stop(self, time_travel: EventLoopClockAdvancer) -> None:
        """Test auto stop functionality."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverAutoStop",
            group_address_long="1/2/1",
            group_address_stop="1/2/2",
            travel_time_down=10,
            travel_time_up=10,
        )
        with patch("time.time") as mock_time:
            mock_time.return_value = 1517000000.0
            # we start with state 0 - open covers (up) this is assumed immediately
            await cover.set_position(0)
            assert xknx.telegrams.qsize() == 1
            _ = xknx.telegrams.get_nowait()

            await cover.set_position(50)

            await time_travel(1)
            mock_time.return_value = 1517000001.0
            assert xknx.telegrams.qsize() == 1
            telegram1 = xknx.telegrams.get_nowait()
            assert telegram1 == Telegram(
                destination_address=GroupAddress("1/2/1"),
                payload=GroupValueWrite(DPTBinary(True)),
            )

            await time_travel(4)
            mock_time.return_value = 1517000005.0
            assert xknx.telegrams.qsize() == 1
            telegram1 = xknx.telegrams.get_nowait()
            assert telegram1 == Telegram(
                destination_address=GroupAddress("1/2/2"),
                payload=GroupValueWrite(DPTBinary(True)),
            )

    async def test_periodic_update(self, time_travel: EventLoopClockAdvancer) -> None:
        """Test periodic update functionality."""
        xknx = XKNX()
        callback_mock = Mock()
        cover = Cover(
            xknx,
            "TestCoverPeriodicUpdate",
            group_address_long="1/2/1",
            group_address_stop="1/2/2",
            group_address_position="1/2/3",
            group_address_position_state="1/2/4",
            travel_time_down=10,
            travel_time_up=10,
            device_updated_cb=callback_mock,
        )
        with patch("time.time") as mock_time:
            mock_time.return_value = 1517000000.0
            # state telegram updates current position - we are not moving so this is new state - not moving
            telegram = Telegram(
                GroupAddress("1/2/4"), payload=GroupValueWrite(DPTArray(0))
            )
            cover.process(telegram)
            await time_travel(0)
            assert callback_mock.call_count == 1
            callback_mock.reset_mock()
            # move to 50%
            telegram = Telegram(
                GroupAddress("1/2/3"), payload=GroupValueWrite(DPTArray(125))
            )
            cover.process(telegram)
            await time_travel(0)
            assert callback_mock.call_count == 1

            mock_time.return_value = 1517000001.0
            await time_travel(1)
            assert callback_mock.call_count == 2

            # state telegram from bus too early
            mock_time.return_value = 1517000001.6
            await time_travel(0.6)
            assert callback_mock.call_count == 2
            telegram = Telegram(
                GroupAddress("1/2/4"), payload=GroupValueWrite(DPTArray(42))
            )
            cover.process(telegram)
            assert callback_mock.call_count == 3
            # next update 1 second after last received state telegram
            mock_time.return_value = 1517000002.0
            await time_travel(0.4)
            assert callback_mock.call_count == 3
            mock_time.return_value = 1517000002.6
            await time_travel(0.6)
            assert callback_mock.call_count == 4
            # last callback - auto updater is removed
            mock_time.return_value = 1517000005.0
            await time_travel(2.4)
            assert callback_mock.call_count == 5
            assert cover.position_reached()
            assert cover._periodic_update_task.done()

    @patch("xknx.core.TelegramQueue.process_telegram_outgoing", new_callable=AsyncMock)
    async def test_remove_task_cancel(
        self, _outgoing: AsyncMock, xknx_no_interface: XKNX
    ) -> None:
        """Test if tasks are removed correctly when device is removed."""
        xknx = xknx_no_interface
        cover = Cover(
            xknx,
            "TestCoverRemoveTaskCancel",
            group_address_long="1/2/1",
            group_address_stop="1/2/2",
            group_address_position_state="1/2/4",
            travel_time_down=10,
            travel_time_up=10,
        )
        xknx.devices.async_add(cover)
        async with xknx:
            # state telegram updates current position - we are not moving so this is new state - not moving
            telegram = Telegram(
                GroupAddress("1/2/4"), payload=GroupValueWrite(DPTArray(0))
            )
            cover.process(telegram)
            assert cover.current_position() == 0
            await cover.set_position(50)
            assert not cover._periodic_update_task.done()
            assert not cover._auto_stop_task.done()
            xknx.devices.async_remove(cover)
            assert cover._periodic_update_task.done()
            assert cover._auto_stop_task.done()

    #
    # HAS GROUP ADDRESS
    #
    def test_has_group_address(self) -> None:
        """Test sensor has group address."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            "TestCoverHasGroupAddress",
            group_address_long="1/2/1",
            group_address_short="1/2/2",
            group_address_position="1/2/3",
            group_address_position_state="1/2/4",
            group_address_angle="1/2/5",
            group_address_angle_state="1/2/6",
        )

        assert cover.has_group_address(GroupAddress("1/2/1"))
        assert cover.has_group_address(GroupAddress("1/2/2"))
        assert cover.has_group_address(GroupAddress("1/2/3"))
        assert cover.has_group_address(GroupAddress("1/2/4"))
        assert cover.has_group_address(GroupAddress("1/2/5"))
        assert cover.has_group_address(GroupAddress("1/2/6"))
        assert not cover.has_group_address(GroupAddress("1/2/7"))



"""
Module for managing a remote temperature value.

DPT 9.001.
"""

from __future__ import annotations

from xknx.dpt import DPTTemperature

from .remote_value import RemoteValue


class RemoteValueTemp(RemoteValue[float]):
    """Abstraction for remote value of KNX 9.001 (DPT_Value_Temp)."""

    dpt_class = DPTTemperature



"""Manages connection callbacks."""

from __future__ import annotations

import asyncio
from collections.abc import Callable
from datetime import datetime
from functools import partial
import logging

from xknx.core.connection_state import XknxConnectionState, XknxConnectionType
from xknx.typing import ConnectionChangeCallbackType

logger = logging.getLogger("xknx.log")


class ConnectionManager:
    """Manages connection state changes XKNX."""

    __slots__ = (
        "_connection_state_changed_cbs",
        "_main_loop",
        "_state",
        "cemi_count_incoming",
        "cemi_count_incoming_error",
        "cemi_count_outgoing",
        "cemi_count_outgoing_error",
        "connected",
        "connected_since",
        "connection_type",
        "undecoded_data_secure",
    )

    def __init__(self) -> None:
        """Initialize ConnectionState class."""
        self._main_loop: asyncio.AbstractEventLoop | None = None

        self.connected = asyncio.Event()
        self._state = XknxConnectionState.DISCONNECTED
        self._connection_state_changed_cbs: list[ConnectionChangeCallbackType] = []

        self.cemi_count_incoming: int = 0
        self.cemi_count_incoming_error: int = 0
        self.cemi_count_outgoing: int = 0
        self.cemi_count_outgoing_error: int = 0
        self.undecoded_data_secure: int = 0
        self.connected_since: datetime | None = None
        self.connection_type: XknxConnectionType = XknxConnectionType.NOT_CONNECTED

    async def register_loop(self) -> None:
        """Register main loop to enable thread-safe `connection_state_changed` calls."""
        self._main_loop = asyncio.get_running_loop()

    def register_connection_state_changed_cb(
        self, connection_state_changed_cb: ConnectionChangeCallbackType
    ) -> Callable[[], None]:
        """Register callback for connection state being updated. Returns unregister function."""
        self._connection_state_changed_cbs.append(connection_state_changed_cb)
        return partial(
            self.unregister_connection_state_changed_cb, connection_state_changed_cb
        )

    def unregister_connection_state_changed_cb(
        self, connection_state_changed_cb: ConnectionChangeCallbackType
    ) -> None:
        """Unregister callback for connection state being updated."""
        if connection_state_changed_cb in self._connection_state_changed_cbs:
            self._connection_state_changed_cbs.remove(connection_state_changed_cb)

    def connection_state_changed(
        self,
        state: XknxConnectionState,
        connection_type: XknxConnectionType = XknxConnectionType.NOT_CONNECTED,
    ) -> None:
        """Run registered callbacks in main loop. Set internal state flag."""
        if self._main_loop:
            self._main_loop.call_soon_threadsafe(
                self._connection_state_changed, state, connection_type
            )
        else:
            self._connection_state_changed(state, connection_type)

    def _connection_state_changed(
        self, state: XknxConnectionState, connection_type: XknxConnectionType
    ) -> None:
        """Run registered callbacks. Set internal state flag."""
        if self._state == state:
            return

        self._state = state
        self.connection_type = connection_type
        if state == XknxConnectionState.CONNECTED:
            self.connected.set()
            self._reset_counters()
        else:
            self.connected.clear()
            self.connected_since = None

        for connection_state_change_cb in self._connection_state_changed_cbs:
            connection_state_change_cb(state)

    @property
    def state(self) -> XknxConnectionState:
        """Get current state."""
        return self._state

    def _reset_counters(self) -> None:
        """Reset counters."""
        self.cemi_count_incoming = 0
        self.cemi_count_incoming_error = 0
        self.cemi_count_outgoing = 0
        self.cemi_count_outgoing_error = 0
        self.undecoded_data_secure = 0
        self.connected_since = datetime.now().astimezone()



"""Unit test for KNX/IP SessionRequest objects."""

from xknx.knxip import HPAI, KNXIPFrame, SessionRequest
from xknx.knxip.knxip_enum import HostProtocol


class TestKNXIPSessionRequest:
    """Test class for KNX/IP SessionRequest objects."""

    def test_session_request(self) -> None:
        """Test parsing and streaming session request KNX/IP packet."""
        public_key = bytes.fromhex(
            "0a a2 27 b4 fd 7a 32 31"  # Diffie-Hellman Client Public Value X
            "9b a9 96 0a c0 36 ce 0e"
            "5c 45 07 b5 ae 55 16 1f"
            "10 78 b1 dc fb 3c b6 31"
        )
        raw = (
            bytes.fromhex(
                "06 10 09 51 00 2e"  # KNXnet/IP header
                "08 02 00 00 00 00 00 00"  # HPAI TCPv4
            )
            + public_key
        )
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, SessionRequest)
        assert knxipframe.body.control_endpoint == HPAI(protocol=HostProtocol.IPV4_TCP)
        assert knxipframe.body.ecdh_client_public_key == public_key

        assert knxipframe.to_knx() == raw

        session_request = SessionRequest(ecdh_client_public_key=public_key)
        knxipframe2 = KNXIPFrame.init_from_body(session_request)

        assert knxipframe2.to_knx() == raw



"""Group address data point type table."""

from __future__ import annotations

from collections.abc import Mapping
import logging

from xknx.dpt.dpt import DPTBase
from xknx.exceptions import ConversionError, CouldNotParseAddress, CouldNotParseTelegram
from xknx.telegram import Telegram, TelegramDecodedData
from xknx.telegram.address import (
    DeviceAddressableType,
    DeviceGroupAddress,
    GroupAddress,
    InternalGroupAddress,
    parse_device_group_address,
)
from xknx.telegram.apci import GroupValueResponse, GroupValueWrite
from xknx.typing import DPTParsable

_GA_DPT_LOGGER = logging.getLogger("xknx.ga_dpt")


class GroupAddressDPT:
    """Class for mapping group addresses to data point types for eager decoding."""

    __slots__ = ("_ga_dpts", "ga_decoding_error")

    def __init__(self) -> None:
        """Initialize GADataTypes class."""
        # using dict[int | str] instead of dict[DeviceGroupAddress] is faster.
        self._ga_dpts: dict[int | str, type[DPTBase]] = {}
        self.ga_decoding_error: set[GroupAddress | InternalGroupAddress] = set()

    def set(
        self,
        ga_dpt: Mapping[DeviceAddressableType, DPTParsable],
    ) -> None:
        """Assign decoders to group addresses."""
        unknown_dpts = set()
        for addr, dpt in ga_dpt.items():
            try:
                address = parse_device_group_address(addr)
            except CouldNotParseAddress as err:
                _GA_DPT_LOGGER.warning("Invalid group address %s: %s", addr, err)
                continue
            if (transcoder := DPTBase.parse_transcoder(dpt)) is None:
                unknown_dpts.add(repr(dpt))  # prevent unhashable types (dict)
                continue
            self._ga_dpts[address.raw] = transcoder
        if unknown_dpts:
            _GA_DPT_LOGGER.debug("No transcoder found for DPTs: %s", unknown_dpts)

    def get(self, address: DeviceGroupAddress) -> type[DPTBase] | None:
        """Return transcoder for group address."""
        return self._ga_dpts.get(address.raw)

    def clear(self) -> None:
        """Clear all group addresses."""
        self._ga_dpts = {}

    def set_decoded_data(self, telegram: Telegram) -> None:
        """Update telegram data with decoded value."""
        if telegram.decoded_data is not None:
            return
        if not isinstance(telegram.payload, GroupValueWrite | GroupValueResponse):
            return
        assert isinstance(  # GroupValueWrite and GroupValueResponse can not have IndividualAddress
            telegram.destination_address, GroupAddress | InternalGroupAddress
        )
        if (transcoder := self.get(telegram.destination_address)) is None:
            return
        try:
            value = transcoder.from_knx(telegram.payload.value)
        except (CouldNotParseTelegram, ConversionError) as err:
            if telegram.destination_address in self.ga_decoding_error:
                _logger_fn = _GA_DPT_LOGGER.debug
            else:
                _logger_fn = _GA_DPT_LOGGER.warning
            self.ga_decoding_error.add(telegram.destination_address)
            _logger_fn(
                "DPT decoding error. Telegram from %s to %s with payload %s can't be decoded by %s: %s",
                telegram.source_address,
                telegram.destination_address,
                telegram.payload.value,
                transcoder.dpt_name(),
                err,
            )
            return
        telegram.decoded_data = TelegramDecodedData(transcoder, value)
